unit FileUtils;

interface

	uses
{$IFC undefined THINK_PASCAL}
		Files, Devices, StandardFile, ToolUtils, 
{$ELSEC}
		InterfacesUI, 
{$ENDC}
		Assertions;

	const
		kTab = Chr(9);
		kReturn = Chr(13);
		kEnter = Chr(3);
		kPeriod = Chr(46);
		kEscape = Chr(27);
		kSpace = Chr(32);
		kComma = Chr(44);

	function GetFullPath (f: FSSpec): str255;

{ Cette fonction appelle StandardGetFile ou StandardPutFile selon que le }
{paramtre 'nouveauFichier' est false ou true. Dans le cas d'un nouveau }
{ fichier, s'il existe dj un fichier de mme nom et que l'utilisateur a }
{ demand  ce qu'il soit remplac, ce fichier est effac par la fonction. }
{ Dans le cas d'un nouveau fichier, les paramtres 'numTypes' et 'typeList' }
{ sont ignors. Dans le cas contraire, l'incitation (prompt) et nom par dfaut }
{ sont ignors. }

	function ChoisisFichier (var reply: StandardFileReply;
									nouveauFichier: Boolean;
									numTypes: Integer;
									typeList: ConstSFTypeListPtr;
									incitation, nom: Str255): Boolean;

	function GetCurrentDateString (format: DateForm): Str255;
	function GetCurrentTimeString (wantSeconds: Boolean): Str255;

{***************************************************}
{*                                                  }
{*   These three functions were created             }
{*   by Greg Ewing (greg@cosc.canterbury.ac.nz,     }
{*   http://www.cosc.canterbury.ac.nz/~greg)        }
{*   November 1996                                  }
{*   Freeware and Use-at-your-own-risk-ware.        }
{*                                                  }
{***************************************************}

{       FSpReset opens an existing file for reading as Pascal text file, given an FSSpec.}
	procedure FSpReset (var f: text;
									spec: FSSpec);

{       FSpAppend opens an existing file for writing as Pascal text file; the text being }
{       appended at the end. }
{ Based on FSpReset by Philippe Casgrain }
	procedure FSpAppend (var f: text;
									spec: FSSpec);

{       FSpRewrite creates a new file with the given type and creator codes, and }
{       opens it for writing as a Pascal text file. }
	procedure FSpRewrite (var f: text;
									spec: FSSpec;
									fileType, fileCreator: OSType);

{       XIOResult may be called to find out the result of the last call to any of the above }
{       routines. Returns noErr if the call succeeded; otherwise it may return either an }
{       Operating System error code, or a return value from IOResult. }
	function XIOResult: OSErr;

implementation

	var
		gXIOResult: OSErr;

	procedure Ignore (x: univ longint);
	begin
{$UNUSED x}
   { procedure to ignore function return values }
	end;

	procedure FSpReset (var f: text;
									spec: FSSpec);

		var
			oldVRefNum: integer;

		procedure Check (result: OSErr);

		begin
			if result <> noErr then begin
				gXIOResult := result;
				exit(FSpReset);
			end;
		end;

	begin {FSpReset}
		gXIOResult := noErr;
		Ignore(GetVol(nil, oldVRefNum));
		Check(HSetVol(nil, spec.vRefNum, spec.parID));
		reset(f, spec.name);
		Check(IOResult);
		Ignore(SetVol(nil, oldVRefNum));
	end;

	procedure FSpRewrite (var f: text;
									spec: FSSpec;
									fileType, fileCreator: OSType);

		var
			oldVRefNum: integer;
			info: FInfo;

		procedure Check (result: OSErr);

		begin
			if result <> noErr then begin
				gXIOResult := result;
				exit(FSpRewrite);
			end;
		end;

	begin
		gXIOResult := noErr;
		Ignore(GetVol(nil, oldVRefNum));
		Ignore(HSetVol(nil, spec.vRefNum, spec.parID));
		rewrite(f, spec.name);
		Check(IOResult);
		Check(FSpGetFInfo(spec, info));
		info.fdType := fileType;
		info.fdCreator := fileCreator;
		Check(FSpSetFInfo(spec, info));
		Ignore(SetVol(nil, oldVRefNum));
	end;

	function XIOResult: OSErr;

	begin
		XIOResult := gXIOResult;
	end;


	procedure FSpAppend (var f: text;
									spec: FSSpec);

		var
			oldVRefNum: integer;

		procedure Check (result: OSErr);

		begin
			if result <> noErr then begin
				gXIOResult := result;
				exit(FSpAppend);
			end;
		end;

	begin {FSpAppend}
		gXIOResult := noErr;
		Ignore(GetVol(nil, oldVRefNum));
		Check(HSetVol(nil, spec.vRefNum, spec.parID));
		Open(f, spec.name);
		Check(IOResult);
		Seek(f, MaxLongInt);
		Ignore(SetVol(nil, oldVRefNum));
	end;

	function GetFullPath;
		var
			tempName, thePath: Str255;
			vParams: CInfoPBRec;
			theError, theErr: OSerr;
			pb: HParamBlockRec;
	begin
		thePath := '';
		tempName := '';
		with vParams do begin
			ioCompletion := nil;
			ioNamePtr := @tempName;
			ioVRefNum := f.vRefNum;
			ioFDirIndex := -1;
			ioDrDirID := f.parID;
			repeat
				theError := PBGetCatInfoSync(@vParams);
				if (theError = noErr) then begin
					pb.ioNamePtr := @tempname;
					pb.ioVRefnum := LoWord(f.parID);
					pb.ioFileID := ioDRParID;
					theErr := PBResolveFileIDRefSync(@pb);
					if theErr = noErr then
						ioDRDirID := pb.ioSrcDirID
					else
						ioDRDirID := ioDRParID;
					thePath := concat(tempName, ':', thePath);
					tempName := '';
				end;
			until (theError <> noErr);
			tempName := concat(thePath, f.name);
			GetFullPath := tempName;
		end;
	end;{ function GetFullPath }

	function ChoisisFichier;
		var
			result: Boolean;
	begin
		if nouveauFichier then begin { StandardPutFile }
			StandardPutFile(incitation, nom, reply);
			result := reply.sfGood;
			if result then
				if reply.sfReplacing then
					result := FSpDelete(reply.sfFile) = noErr;
		end
		else begin { StandardGetFile }

			if numTypes > 0 then
				Assert(typeList <> nil);

{$IFC undefined THINK_PASCAL}
			StandardGetFile(nil, numTypes, typeList, reply);
{$ELSEC}
			StandardGetFile(nil, numTypes, typeList^, reply);
{$ENDC}

			result := reply.sfGood;
		end;

		ChoisisFichier := result;

	end; { ChoisisFichier }

	function GetCurrentDateString (format: DateForm): Str255;
		var
			s: Str255;
			secs: LongInt;
	begin
		GetDateTime(secs);
		IUDateString(secs, format, s);
		GetCurrentDateString := s;
	end; { GetCurrentDateString }

	function GetCurrentTimeString (wantSeconds: Boolean): Str255;
		var
			s: Str255;
			secs: LongInt;
	begin
		GetDateTime(secs);
		IUTimeString(secs, wantSeconds, s);
		GetCurrentTimeString := s;
	end; { GetCurrentTimeString }

end.