unit MatrixAlgebra;

{**********}
interface
{**********}

	uses
{$IFC undefined THINK_PASCAL}
		Memory, 
{$endc}
		Assertions, MathDeclarations, Permutations, AlertUtils;


{ Note sur l'inversion matricielle (25/06/98), sur une suggestion de PL }
{ J'ai adapt la procdure GAUSSJ de Numerical Recipes (Press et al., 1986) }
{ pour remplacer ma propre procdure d'unversion qui est (beaucoup) plus lente }
{ et surtout perd de la prcision plus rapidement. Certaines valeurs devenaient }
{ mme des NANs alors que cela n'tait pas ncessaire. }
{ J'ai donc install ci-dessous une directive de compilation: mettez-l  true }
{ pour obtenir l'inversion avec GaussJ et false pour l'ancienne procdure (qui }
{ tait d'ailleurs copie verbatim de cologie Numrique, 2nde dition (1983)) }
{ et qui utilisait le calcul des co-facteurs, donc pas mal plus lent. }

{$setc use_gaussj_inversion := true}



{ Faire des copies de matrices }
{ Cette fonction copie intgralement une matrice, et peut omettre une ligne/colonne }
	procedure DuplicateMatrix (source: TMatrixPtr;
									var result: TMatrixPtr;
									excludeRow, excludeColumn: TInteger;
									withNames: Boolean);
{ Cette fonction copie une ou des colonnes de la matrice source en la matrice rsultats }
	procedure SubMatrix (source: TMatrixPtr;
									var result: TMatrixPtr;
									whatToCopy: TBoolVectorPtr;
									withNames: Boolean);

{ Oprations sur les matrices }
	function MDet (mat: TMatrixPtr): TReal;
	function MInverse (var mat: TMatrixPtr): Boolean;
	procedure MTranspose (mat: TMatrixPtr;
									var result: TMatrixPtr);
{ Dans cette procdure j'a rajout trois paramtres d'indirection, car il est parfois }
{ ncessaires de faire une indirection sur la matrice c, comme dans le cas de tests }
{ par permutations. Voir ElementPermute. }
	procedure MMult (b, c: TMatrixPtr;
									var result: TMatrixPtr;
									transposerB: Boolean);

{ Procdure tire de Numerical Recipes pour trouver les valeurs propres d'une matrice }
{ carre asymtrique. Cette dernire est d'ailleurs modifie par la fonction. }
	function EigenValues (var b: TMatrixPtr): TMatrixPtr;

{****************}
implementation
{****************}

	procedure DuplicateMatrix (source: TMatrixPtr;
									var result: TMatrixPtr;
									excludeRow, excludeColumn: TInteger;
									withNames: Boolean);
		var
			i, j, oneLessRow, oneLessColumn, rowCount, colCount: TInteger;
	begin
		Assert(source <> nil);
		if source <> nil then begin
{ User may wish to omit one row/column from the duplication, e.g. Cofactor matrix }
			if excludeRow > 0 then
				oneLessRow := 1
			else
				oneLessRow := 0;
			if excludeColumn > 0 then
				oneLessColumn := 1
			else
				oneLessColumn := 0;

			if not NewMatrix(result, source^.n - oneLessRow, source^.p - oneLessColumn, withNames, withNames, source^.permMeth) then
				ProblemAlert(errNoMatrixMemory, 0, 'DuplicateMatrix', '');
			if (source^.col <> nil) then begin { Copier les donnes des colonnes }
				colCount := 0;
				if withNames and (source^.rowName <> nil) then begin { copier les noms de lignes }
					rowCount := 0;

					if source^.permMeth > permVector then { tenir compte du petit nombre de noms si matrice de distance }
						j := source^.nbObjets
					else
						j := source^.n;

					for i := 1 to j do
						if (i <> excludeRow) then begin { can copy this row name }
							rowCount := rowCount + 1;
							result^.rowName^[rowCount] := source^.rowName^[i];
						end; { for i }
				end; { if }

				for j := 1 to source^.p do
					if (j <> excludeColumn) then begin { can copy this column }
						colCount := colCount + 1;
						if withNames and (source^.colName <> nil) then
							result^.colName^[colCount] := source^.colName^[j];
						result^.col^[colCount].n := source^.col^[j].n;
						rowCount := 0;
						for i := 1 to source^.n do
							if (i <> excludeRow) then begin { can copy this row element }
								rowCount := rowCount + 1;
								result^.col^[colCount].data^[rowCount] := source^.col^[j].data^[i];
							end; { for i }
					end; { for j }
			end; { if source^.col }
			if (excludeRow = 0) and (excludeColumn = 0) then begin { Copier ceci seulement quand on n'exclut rien }
				result^.root := source^.root;
				if result^.permMeth <> permNone then
					if source^.indirection <> nil then begin { Copier les donnes d'indirection }
						CreerVecteurIndirection(result);
						if result^.permMeth <= permVector then
							rowCount := result^.n
						else
							rowCount := result^.nbObjets;
						for i := 1 to rowCount do
							result^.indirection^[i] := source^.indirection^[i];
					end; { if indirection }

				if source^.group <> nil then begin { copier les groupes }
					CreerVecteurGroupes(result);
					if result^.permMeth <= permVector then
						rowCount := result^.n
					else
						rowCount := result^.nbObjets;
					for i := 1 to rowCount do
						result^.group^[i] := source^.group^[i];
				end; { if source^.group }

				if source^.include <> nil then begin { copier les inclusions }
					if result^.permMeth <= permVector then
						rowCount := result^.n
					else
						rowCount := result^.nbObjets;
					for i := 1 to rowCount do
						result^.include^[i] := source^.include^[i];
				end; { if source^.include }

			end; { if nothing excluded }
		end; { if source <> nil }
	end; { DuplicateMatrix }

	procedure SubMatrix (source: TMatrixPtr;
									var result: TMatrixPtr;
									whatToCopy: TBoolVectorPtr;
									withNames: Boolean);
		var
			colCount, i, j: TInteger;
	begin
		Assert(source <> nil);
		if source <> nil then begin
{ Trouver le nombre de colonnes  copier }
			colCount := 0;
			if source^.p > 0 then
				for j := 1 to source^.p do
					if whatToCopy^[j] then
						colCount := colCount + 1;
			if not NewMatrix(result, source^.n, colCount, withNames, withNames, source^.permMeth) then
				ProblemAlert(errNoMatrixMemory, 0, 'SubMatrix', '');
			if (source^.col <> nil) and (result^.col <> nil) then begin
				colCount := 0;
				if withNames and (source^.rowName <> nil) then
					for i := 1 to source^.n do
						result^.rowName^[i] := source^.rowName^[i];
				for j := 1 to source^.p do
					if whatToCopy^[j] then begin
						colCount := colCount + 1;
						if withNames and (source^.colName <> nil) then
							result^.colName^[colCount] := source^.colName^[j];
						result^.col^[colCount].n := source^.col^[j].n;
						for i := 1 to source^.n do
							result^.col^[colCount].data^[i] := source^.col^[j].data^[i];
					end; { for j }
			end; { if source^.col }
		end; { if source <> nil }
	end; { SubMatrix }

	function MDet (mat: TMatrixPtr): TReal;
		var
			det, pivot, val: TReal;
			i, j, k: TInteger;
			tMat: TMatrixPtr;
	begin
		det := ERR_RESULT;
		tMat := nil;
		Assert(mat <> nil);
		if mat <> nil then
			if (mat^.n = mat^.p) then { matrice carre, parfait }
				if mat^.n > 1 then begin
					DuplicateMatrix(mat, tMat, 0, 0, false); {fill tMat with mat, no need to omit a row/col}
					if tMat <> nil then
						for k := 1 to tMat^.n - 1 do
							for i := k + 1 to tMat^.n do {condensation pivotale pour la k-ime colonne}
								begin
								pivot := GetElement(tMat, i, k) / GetElement(tMat, k, k);
								for j := 1 to tMat^.n do begin
									val := GetElement(tMat, i, j) - (pivot * GetElement(tMat, k, j));
									SetElement(tMat, i, j, val);
								end;
							end;
					det := GetElement(tMat, 1, 1);
					for i := 2 to tMat^.n do
						det := det * GetElement(tMat, i, i);
					DisposeMatrix(tMat);
				end {if n > 1}
				else
					det := GetElement(mat, 1, 1);
		MDet := det;
	end; {MDet}

{$ifc use_gaussj_inversion}

{ This is procedure GAUSSJ from Numerical Recipes, here named MINVERSE }
{ Adapted for dynamically-sized matrices by Philippe Casgrain }
{ 25/06/98 }
{ Matrix B is a zero-filled, nx1 vector. On output it may contain the linear regression }
{ coefficients solution vector, but here we don't really care about that, just the inverse. }
{ The function returns FALSE if mat is a singular matrix (that cannot be inverted) }

	function MInverse (var mat: TMatrixPtr): Boolean;
		var
			n, m: TInteger;
			b: TMatrixPtr;
			big, dum, pivinv: TReal;
			i, icol, irow, j, k, l, ll: TInteger;
			indxc, indxr, ipiv: TIntegerVectorPtr;
			result: Boolean;
	begin
		Assert(mat <> nil);
{ Init variables }
		result := true;
		n := mat^.n;
		m := 1;
{ Create the matrices and vectors }
		b := nil;
		if NewMatrix(b, n, m, false, false, mat^.permMeth) then
			;
		indxc := NewTIntegerVectorPtr(n);
		indxr := NewTIntegerVectorPtr(n);
		ipiv := NewTIntegerVectorPtr(n);

		if (b = nil) or (indxc = nil) or (indxr = nil) or (ipiv = nil) then begin
			result := false;
			ProblemAlert(errNoMatrixMemory, 0, 'GaussJ inversion', '');
		end;

		if result then begin
			for j := 1 to n do begin
				ipiv^[j] := 0
			end;
			for i := 1 to n do begin
				big := 0.0;
				for j := 1 to n do begin
					if (ipiv^[j] <> 1) then begin
						for k := 1 to n do begin
							if (ipiv^[k] = 0) then begin
								if (abs(GetElement(mat, j, k)) >= big) then begin
									big := abs(GetElement(mat, j, k));
									irow := j;
									icol := k;
								end
							end
							else if (ipiv^[k] > 1) then begin
								result := false; {pause 1 in GAUSSJ - singular matrix}
							end
						end
					end
				end;
				ipiv^[icol] := ipiv^[icol] + 1;
				if (irow <> icol) then begin
					for l := 1 to n do begin
						dum := GetElement(mat, irow, l);
						SetElement(mat, irow, l, GetElement(mat, icol, l));
						SetElement(mat, icol, l, dum)
					end;
					for l := 1 to m do begin
						dum := GetElement(b, irow, l);
						SetElement(b, irow, l, GetElement(b, icol, l));
						SetElement(b, icol, l, dum)
					end
				end;
				indxr^[i] := irow;
				indxc^[i] := icol;
				if (GetElement(mat, icol, icol) = 0.0) then begin
					result := false; {pause 2 in GAUSSJ - singular matrix}
				end;
				pivinv := 1.0 / GetElement(mat, icol, icol);
				SetElement(mat, icol, icol, 1.0);
				for l := 1 to n do begin
					SetElement(mat, icol, l, GetElement(mat, icol, l) * pivinv)
				end;
				for l := 1 to m do begin
					SetElement(b, icol, l, GetElement(b, icol, l) * pivinv)
				end;
				for ll := 1 to n do begin
					if (ll <> icol) then begin
						dum := GetElement(mat, ll, icol);
						SetElement(mat, ll, icol, 0.0);
						for l := 1 to n do begin
							SetElement(mat, ll, l, GetElement(mat, ll, l) - GetElement(mat, icol, l) * dum)
						end;
						for l := 1 to m do begin
							SetElement(b, ll, l, GetElement(b, ll, l) - GetElement(b, icol, l) * dum)
						end
					end
				end
			end;
			for l := n downto 1 do begin
				if (indxr^[l] <> indxc^[l]) then begin
					for k := 1 to n do begin
						dum := GetElement(mat, k, indxr^[l]);
						SetElement(mat, k, indxr^[l], GetElement(mat, k, indxc^[l]));
						SetElement(mat, k, indxc^[l], dum)
					end
				end
			end;
		end; { if result }

{ Dispose of matrix and vectors }
		DisposeMatrix(b);
		if (indxc <> nil) then
			DisposePtr(Ptr(indxc));
		if (indxr <> nil) then
			DisposePtr(Ptr(indxr));
		if (ipiv <> nil) then
			DisposePtr(Ptr(ipiv));

		MInverse := result;
	end; { procedure gaussj }

{$elsec}

	function MCofact (mat: TMatrixPtr;
									line, col: TInteger): TReal;
		var
			cof: TReal;
			signe: TInteger;
			tMat: TMatrixPtr;
	begin
		cof := ERR_RESULT;
		tMat := nil;
		Assert(mat <> nil);
		if mat <> nil then
			if (mat^.n = 2) then begin
				signe := ((-1) * ((line + col) mod 2));
				if (signe = 0) then
					signe := 1;
				cof := signe * GetElement(mat, (mat^.n + 1) - line, (mat^.n + 1) - col);
			end
			else {calculer le cofacteur pour vrai!}
				begin
{fill tMat with mat, excluding one row and one column}
				DuplicateMatrix(mat, tMat, line, col, false);

				signe := ((-1) * ((line + col) mod 2));
				if (signe = 0) then
					signe := 1;
				cof := signe * MDet(tMat);
				DisposeMatrix(tMat);
			end;
		MCofact := cof;
	end;{MCofact}

	function MInverse (var mat: TMatrixPtr): Boolean;
		var
			tMat: TMatrixPtr;
			i, j: TInteger;
			det, value: TReal;
			result: Boolean;
	begin
		result := false; { Inversion encore non-russie }
		tMat := nil;
		Assert(mat <> nil);
		if mat <> nil then
			if (mat^.n = mat^.p) then { Matrice carre, on peut continuer }
				if mat^.n > 1 then begin
					DuplicateMatrix(mat, tMat, 0, 0, false); {fill tMat with mat, no need to omit a row/col}
					det := MDet(tMat);
					if det = 0 then
						result := false {matrice singulire, non-inversible}
					else begin {dbut inversion}
						result := true;
						det := 1 / det;
						for i := 1 to mat^.n do
							for j := 1 to mat^.n do begin
								value := det * MCofact(tMat, j, i);
								SetElement(mat, i, j, value);
							end;
					end; {inversion}
					DisposeMatrix(tMat);
				end {if n > 1}
				else if mat^.n = 1 then begin
					value := GetElement(mat, 1, 1);
					if (value <> 0) then begin
						value := 1 / value;
						SetElement(mat, 1, 1, value);
						result := true;
					end {if mat^[1, 1] <> 0}
					else
						result := false {car le seul element de la matrice est nul}
				end
				else
					result := false; {car n < 1!!!}

		MInverse := result;
	end;{MInverse}

{$endc}

	procedure MTranspose (mat: TMatrixPtr;
									var result: TMatrixPtr);
		var
			i, j: TInteger;
			valIJ: TReal;
	begin
		Assert(mat <> nil);
		if mat <> nil then
			if NewMatrix(result, mat^.p, mat^.n, false, false, mat^.permMeth) then begin
				for i := 1 to mat^.n do
					for j := 1 to mat^.p do begin
						valIJ := GetElement(mat, i, j);
						SetElement(result, j, i, valIJ); { Transposition }
					end;
			end; { if NewMatrix }
	end; { MTranspose }

	procedure MMult (b, c: TMatrixPtr;
									var result: TMatrixPtr;
									transposerB: Boolean);
		var
			i, j, k, nLignes, compteColonnes: TInteger;
			sum, bb, cc: TReal;
	begin
		Assert((b <> nil) and (c <> nil));
		if (b <> nil) and (c <> nil) then begin
{ Taille de la matrice result }
			if transposerB then begin
				nLignes := b^.p;
				compteColonnes := b^.n; { l'autre dimension }
			end
			else begin
				nLignes := b^.n;
				compteColonnes := b^.p;
			end;
{ Condition essentielle  la multiplication: nombre de colonnes de la matrice b doit }
{ tre gal au nombre de lignes de c. }
			if (compteColonnes = c^.n) then
				if NewMatrix(result, nLignes, c^.p, false, false, b^.permMeth) then begin
					for j := 1 to c^.p do begin
						for i := 1 to nLignes do begin
							sum := 0;
							for k := 1 to compteColonnes do begin
								if transposerB then
									bb := GetElement(b, k, i)
								else
									bb := GetElement(b, i, k);
{ 29/01/98 PhC Rajout de ElementPermute pour pouvoir permuter les lignes de la matrice c }
{ 30/04/98 PhC ajout d'un test pour savoir s'il existe un vecteur d'indirection }
{ pour permuter, sinon on calcule normalement }
								if c^.indirection = nil then
									cc := GetElement(c, k, j)
								else
									cc := GetElement(c, ElementPermute(k, c), j);

								sum := sum + bb * cc;
							end; { k }
							SetElement(result, i, j, sum);
						end; { i }
					end; { j }
				end; { if NewMatrix }
		end; { if <>nil }
	end; { MMult }

{$ifc false}
	procedure InitFactSum;
		var
			i: Integer;
	begin
		factSum^[1] := nbObjets - 1;
		for i := 2 to nbObjets do
			factSum^[i] := factSum^[i - 1] + nbObjets - i;
	end;

	function MtS (i, j: Integer): Integer;
		var
			temp: Integer;
	begin
		if i > j then {echanger les valeurs de i et j: triangulaire superieure}
			begin
			temp := i;
			i := j;
			j := temp;
		end;{if i > j}
{ A verifier... }
		if i = j then
			MtS := factSum^[i] - (nbObjets - j) + 1
		else
			MtS := factSum^[i] - (nbObjets - j);
	end;

	procedure StM (a: Integer;
									var i, j: Integer);
	begin
		i := 0;
		repeat
			i := i + 1
		until factSum^[i] >= a;
		j := nbObjets - (factSum^[i] - a);
	end;

{$endc}

	function EigenValues;

{ Three procedures from Numerical Recipes for eigenvalues of asymmetric matrices }
		procedure balanc (var a: TMatrixPtr);
(* Programs using routine BALANC should define the type *)
(*    type *)
(*    glnpnp = array[1..np, 1..np] of real; *)
(*    where 'np by np' is the physical dimension of the *)
(*    array to be analyzed . *)
			const
				radix = 2.0;
			var
				last, j, i: TInteger;
				s, r, g, f, c, sqrdx: TReal;
		begin
			sqrdx := sqr(radix);
			repeat
				last := 1;
				for i := 1 to a^.n do begin
					c := 0.0;
					r := 0.0;
					for j := 1 to a^.n do
						if (j <> i) then begin
							c := c + abs(GetElement(a, j, i));
							r := r + abs(GetElement(a, i, j))
						end;
					if ((c <> 0.0) and (r <> 0.0)) then begin
						g := r / radix;
						f := 1.0;
						s := c + r;
						while (c < g) do begin
							f := f * radix;
							c := c * sqrdx
						end;
						g := r * radix;
						while (c > g) do begin
							f := f / radix;
							c := c / sqrdx
						end;
						if ((c + r) / f < 0.95 * s) then begin
							last := 0;
							g := 1.0 / f;
							for j := 1 to a^.n do
								SetElement(a, i, j, GetElement(a, i, j) * g);
							for j := 1 to a^.n do
								SetElement(a, j, i, GetElement(a, j, i) * f)
						end
					end
				end;
			until (last <> 0)
		end;

		procedure elmhes (var a: TMatrixPtr);
(* Programs using routine ELMHES must define the type *)
(*    type *)
(*    glnpnp = ARRAY [ 1 .. np , 1 .. np ] where 'np by np' is the physical dimension of the matrix to be reduced . *)
			var
				m, j, i: TInteger;
				y, x: TReal;
		begin
			if (a^.n > 2) then begin
				for m := 2 to a^.n - 1 do begin
					x := 0.0;
					i := m;
					for j := m to a^.n do begin
						if (abs(GetElement(a, j, m - 1)) > abs(x)) then begin
							x := GetElement(a, j, m - 1);
							i := j
						end
					end;
					if (i <> m) then begin
						for j := m - 1 to a^.n do begin
							y := GetElement(a, i, j);
							SetElement(a, i, j, GetElement(a, m, j));
							SetElement(a, m, j, y);
						end;
						for j := 1 to a^.n do begin
							y := GetElement(a, j, i);
							SetElement(a, j, i, GetElement(a, j, m));
							SetElement(a, j, m, y);
						end
					end;
					if (x <> 0.0) then begin
						for i := m + 1 to a^.n do begin
							y := GetElement(a, i, m - 1);
							if (y <> 0.0) then begin
								y := y / x;
								SetElement(a, i, m - 1, y);
								for j := m to a^.n do
									SetElement(a, i, j, GetElement(a, i, j) - y * GetElement(a, m, j));
								for j := 1 to a^.n do
									SetElement(a, j, m, GetElement(a, j, m) + y * GetElement(a, j, i))
							end
						end
					end
				end
			end
		end;

		procedure hqr (var a: TMatrixPtr;
										var wrwi: TMatrixPtr); { Mod. PhC: wrwi is a nx2 matrix instead of being 2 nx1 vectors. real = col 1 and imaginary = col 2 }
(* Programs using routine HQR must define the type *)
(*    type *)
(*    glnpnp = array[1..np, 1..np] of real; *)
(*    glnp = array[1..np] of real; *)
(*    where 'np by np' is the physical dimension of the matrix whose eigenvalues are to be found . *)
			label
				2, 3, 4;
			var
				nn, m, l, k, j, its, i, mmin: TInteger;
				z, y, x, w, v, u, t, s, r, q, p, anorm: TReal;
			function sign (a, b: TReal): TReal;
			begin
				if (b < 0.0) then
					sign := -abs(a)
				else
					sign := abs(a)
			end;
			function min (a, b: TInteger): TInteger;
			begin
				if (a < b) then
					min := a
				else
					min := b
			end;
		begin
			anorm := abs(GetElement(a, 1, 1));
			for i := 2 to a^.n do begin
				for j := i - 1 to a^.n do begin
					anorm := anorm + abs(GetElement(a, i, j))
				end
			end;
			nn := a^.n;
			t := 0.0;
			while (nn >= 1) do begin
				its := 0;
2:
				for l := nn downto 2 do begin
					s := abs(GetElement(a, l - 1, l - 1)) + abs(GetElement(a, l, l));
					if (s = 0.0) then
						s := anorm;
					if ((abs(GetElement(a, l, l - 1)) + s) = s) then
						goto 3
				end;
				l := 1;
3:
				x := GetElement(a, nn, nn);
				if (l = nn) then begin
					SetElement(wrwi, nn, 1, x + t);
					SetElement(wrwi, nn, 2, 0.0);
					nn := nn - 1
				end
				else begin
					y := GetElement(a, nn - 1, nn - 1);
					w := GetElement(a, nn, nn - 1) * GetElement(a, nn - 1, nn);
					if (l = nn - 1) then begin
						p := 0.5 * (y - x);
						q := sqr(p) + w;
						z := sqrt(abs(q));
						x := x + t;
						if (q >= 0.0) then begin
							z := p + sign(z, p);
							SetElement(wrwi, nn, 1, x + z);
							SetElement(wrwi, nn - 1, 1, GetElement(wrwi, nn, 1));
							if (z <> 0.0) then
								SetElement(wrwi, nn, 1, x - w / z);
							SetElement(wrwi, nn, 2, 0.0);
							SetElement(wrwi, nn - 1, 2, 0.0);
						end
						else begin
							SetElement(wrwi, nn, 1, x + p);
							SetElement(wrwi, nn - 1, 1, GetElement(wrwi, nn, 1));
							SetElement(wrwi, nn, 2, z);
							SetElement(wrwi, nn - 1, 2, -z)
						end;
						nn := nn - 2
					end
					else begin
						if (its = 30) then begin
							ProblemAlert(errGenericProblem, 0, 'pause in routine HQR: too many iterations', '');
						end;
						if ((its = 10) or (its = 20)) then begin
							t := t + x;
							for i := 1 to nn do begin
								SetElement(a, i, i, GetElement(a, i, i) - x)
							end;
							s := abs(GetElement(a, nn, nn - 1)) + abs(GetElement(a, nn - 1, nn - 2));
							x := 0.75 * s;
							y := x;
							w := -0.4375 * sqr(s)
						end;
						its := its + 1;
						for m := nn - 2 downto l do begin
							z := GetElement(a, m, m);
							r := x - z;
							s := y - z;
							p := (r * s - w) / GetElement(a, m + 1, m) + GetElement(a, m, m + 1);
							q := GetElement(a, m + 1, m + 1) - z - r - s;
							r := GetElement(a, m + 2, m + 1);
							s := abs(p) + abs(q) + abs(r);
							p := p / s;
							q := q / s;
							r := r / s;
							if (m = l) then
								goto 4;
							u := abs(GetElement(a, m, m - 1)) * (abs(q) + abs(r));
							v := abs(p) * (abs(GetElement(a, m - 1, m - 1)) + abs(z) + abs(GetElement(a, m + 1, m + 1)));
							if ((u + v) = v) then
								goto 4
						end;
4:
						for i := m + 2 to nn do begin
							SetElement(a, i, i - 2, 0.0);
							if (i <> (m + 2)) then
								SetElement(a, i, i - 3, 0.0)
						end;
						for k := m to nn - 1 do begin
							if (k <> m) then begin
								p := GetElement(a, k, k - 1);
								q := GetElement(a, k + 1, k - 1);
								r := 0.0;
								if (k <> (nn - 1)) then
									r := GetElement(a, k + 2, k - 1);
								x := abs(p) + abs(q) + abs(r);
								if (x <> 0.0) then begin
									p := p / x;
									q := q / x;
									r := r / x
								end
							end;
							s := sign(sqrt(sqr(p) + sqr(q) + sqr(r)), p);
							if (s <> 0.0) then begin
								if (k = m) then begin
									if (l <> m) then
										SetElement(a, k, k - 1, -GetElement(a, k, k - 1));
								end
								else begin
									SetElement(a, k, k - 1, -s * x)
								end;
								p := p + s;
								x := p / s;
								y := q / s;
								z := r / s;
								q := q / p;
								r := r / p;
								for j := k to nn do begin
									p := GetElement(a, k, j) + q * GetElement(a, k + 1, j);
									if (k <> (nn - 1)) then begin
										p := p + r * GetElement(a, k + 2, j);
										SetElement(a, k + 2, j, GetElement(a, k + 2, j) - p * z)
									end;
									SetElement(a, k + 1, j, GetElement(a, k + 1, j) - p * y);
									SetElement(a, k, j, GetElement(a, k, j) - p * x)
								end;
								mmin := min(nn, k + 3);
								for i := l to mmin do begin
									p := x * GetElement(a, i, k) + y * GetElement(a, i, k + 1);
									if (k <> (nn - 1)) then begin
										p := p + z * GetElement(a, i, k + 2);
										SetElement(a, i, k + 2, GetElement(a, i, k + 2) - p * r)
									end;
									SetElement(a, i, k + 1, GetElement(a, i, k + 1) - p * q);
									SetElement(a, i, k, GetElement(a, i, k) - p)
								end
							end
						end;
						goto 2
					end
				end
			end
		end;


		var
			wrwi: TMatrixPtr;
	begin
		wrwi := nil;
		Assert(b <> nil);
		Assert(b^.n = b^.p);
		if NewMatrix(wrwi, b^.n, 2, false, false, permNone) then begin
			balanc(b);
			elmhes(b);
			hqr(b, wrwi);
		end
		else
			ProblemAlert(errNoMatrixMemory, 0, 'EigenValues', '');

		EigenValues := wrwi;
	end; { EigenValues }

end. {implementation of MatrixAlgebra}