unit Permutations;

{ Version 3.4: complete rewrite using TransSkel 2.6 and PowerPC code }
{              Uses dynamic memory }
{              Simply does permutations now }

interface

	uses
		MathDeclarations, Assertions, AlertUtils;{, QuickDraw}

	const
		INTEGER_ARRAY = True; { Pour rendre SHELLSORT plus lisible }
		REAL_ARRAY = False;

	type
		TPermuteVect = record
				nbVar: TInteger;
				ppNeg: TIntegerVectorPtr;
				pgPos: TIntegerVectorPtr;
				milieu: TIntegerVectorPtr;
				egNeg: TIntegerVectorPtr;
				egPos: TIntegerVectorPtr;
				probB: TRealVectorPtr;
				probR: TReal;
				egr, pgr: TInteger;
			end;

{ Trouver la ligne/colonne de l'lment elem dans la triangulaire suprieure dplie }
	procedure ElementVersIJ (elem, nbObjets: TInteger;
									var i, j: TInteger);
{ Trouver l'lment dans la triangulaire suprieure de la matrice de distance }
{ dplie vers la droite }
	function IJVersElement (nbObjets, i, j: TInteger): TInteger;

{ Cette fonction retourne l'indice d'lment tant donn le vecteur d'indirection (si }
{ non-nul). Si nombreObjets > 0, il s'agit d'une permutation  la Mantel }
	function ElementPermute (elem: TInteger;
									m: TMatrixPtr): TInteger;

	procedure CreatePermuteVect (nombreDeVariables: TInteger;
									var result: TPermuteVect);
	procedure ClearPermuteVect (nombreDeVariables: TInteger;
									var result: TPermuteVect);
	procedure DisposePermuteVect (var result: TPermuteVect);

{ Procdure de tri avec indices associs, utile pour indirections et autres }
	procedure ShellSort (n: integer;
									var randArr: TRealVectorPtr; { le vecteur  trier }
									var indepArr: univ ptr; { vecteur d'entiers ou de rels  entraner }
									isInteger: Boolean);

{ Permuter un vecteur v selon le modle permutationnel choisi }
{ Le vecteur "original" reprsente les valeurs du vecteur dans leur ordre }
{ original, et est ncessaire pour les modles permutationnels "double" et "triple". }
{ Pour les autres, passer "nil". }
	procedure PermuterUnVecteur (var v: TMatrixPtr;
									original: TMatrixPtr;
									permutationsLimitees: Boolean);

	procedure InitRan3 (initSeed, initCount: integer);
	function Ran3: real;

implementation

	var
{ Pour Ran3 de Numerical Recipes }
		glinext, glinextp: integer;
		glma: array[1..55] of real;
		gSeed: integer;

	procedure ElementVersIJ (elem, nbObjets: TInteger;
									var i, j: TInteger);
		var
			c, p: TInteger;
	begin
		i := 0;
		p := 0;
		c := nbObjets;
		while (p < elem) do begin
			c := c - 1;
			p := p + c;
			i := i + 1;
		end; { while }
{ au sortir de cette boucle on a la bonne valeur de i. }
{ Maintenant, calculons la valeur de j: }
		j := nbObjets - (p - elem);
	end; { ElementVersIJ }

	function IJVersElement (nbObjets, i, j: TInteger): TInteger;
		var
			temp: TInteger;
	begin
{                                                                               }
{ Bonjour Philippe.                                 Le 28 fvrier 1996, 1:00 am }
{                                                                               }
{ Voici la formule que tu cherches.                                             }
{                                                                               }
{ n = nombre d'objets dans la matrice carre.                                   }
{ La position de la valeur recherche est (i,j)                                 }
{                                                                               }
{ Position dans le vecteur correspondant  la triangulaire suprieure dplie   }
{ par lignes: j-i+(2n-i)*(i-1)/2                                                }
{                                                                               }
{ Voici comment j'y suis arriv:                                                }
{                                                                               }
{ -> Il y a (i-1) lignes pleines avant celle o se trouve la case (i,j).        }
{                                                                               }
{ -> Ces (i-1) lignes pleines sont constitues de (i-1)*n valeurs MOINS         }
{ celles qui se trouvent sur la diagonale et  gauche de celle-ci.              }
{                                                                               }
{ -> Ces valeurs  soustraire des lignes pleines se calculent par la somme      }
{ des premiers entiers de 1 jusqu' i: il faut enlever une case  la premire   }
{ ligne, deux  la seconde ligne, etc. jusqu' la ligne (i-1). La formule       }
{ pour la somme des premiers entiers se trouve dans Ecologie numrique! (Tome   }
{ 1, p. 156). Pour la somme de 1 jusqu' (i-1), la formule donne i*(i-1)/2.     }
{                                                                               }
{ -> A ceci il faut ajouter le nombre de cases de la dernire ligne: il y en    }
{ a j-i.                                                                        }
{                                                                               }
{ ->  Tu assembles tout a et a te donne: (i-1)*n - i*(i-1)/2 + (j-1), ce      }
{ qui se condense en ma formule du dbut.                                       }
{                                                                               }
{ Amuse-toi bien avec cette formule!                                            }
{                                                                               }

{ Mod. PhC 03/06/98: changer i et j si i>j (tr. inf.) }
		if i > j then begin
			temp := i;
			i := j;
			j := temp;
		end; { if i > j }

		IJVersElement := (i - 1) * nbObjets - i * (i - 1) div 2 + (j - 1) - i + 1;
	end; { IJVersElement }

	function ElementPermute (elem: TInteger;
									m: TMatrixPtr): TInteger;
		var
			i, j, k: TInteger;
	begin
		Assert(m <> nil);

		if m^.indirection <> nil then begin { Mod PhC 03/06/98, excuter seulement si vecteur d'indirection, sinon retourner elem }
			case m^.permMeth of
				permNone: 
					ElementPermute := elem; { Pas de permutation  faire }

				permVector:  begin
{ facile! Le vecteur d'indirection est un long vecteur correspondant au vecteur  permuter }
{ Vrifier d'abord si les conditions sont respectes... }
					if (elem > 0) and (elem <= m^.n) then
						ElementPermute := m^.indirection^[elem]
					else
						MyDebugStr('Error in ElementPermute: element out of range!'); { erreur! }
				end; { permVector }

				permMatrix, permDouble, permTriple:  begin { Mthode de permutation Mantel/double/triple (c'est pareil) }
					Assert(m^.nbObjets > 0);
{ plus compliqu... On s'attend  ce que le vecteur d'indirection reprsente  la }
{ fois les lignes et colonnes d'une matrice de distances dplie (triangulaire suprieure, }
{ sans la diagonale... }

{ Ce qu'on a: une position "elem" dans un vecteur 1..n, reprsentant une matrice de }
{ distances dplie. Il faut donc... }
{ 1) Trouver la position (i,j) du elem-ime lment dans cette matrice (nombreObjets }
{    x nombreObjets). }
					ElementVersIJ(elem, m^.nbObjets, i, j);

					Assert(i <= m^.nbObjets);
					Assert(j <= m^.nbObjets);

{ 2) Convertir ce (i,j) en (i',j')  l'aide du vecteur d'indirection (facile) }

					i := m^.indirection^[i];
					j := m^.indirection^[j];
					if i > j then begin { changer i et j }
						k := i;
						i := j;
						j := k;
					end;
{ i et j sont maintenant i' et j' }

{ 3) Re-convertir ce (i',j') en un lment permut elem', et retourner elem' en rsultat }
					ElementPermute := IJVersElement(m^.nbObjets, i, j);
				end; { permMatrix, double, triple }

				otherwise
					MyDebugStr('Should not get here in ElementPermute!');
			end; { case }

		end
		else
			ElementPermute := elem;

	end; { ElementPermute }

	procedure CreatePermuteVect (nombreDeVariables: TInteger;
									var result: TPermuteVect);
	begin
		result.nbVar := nombreDeVariables;
		with result do begin
			ppNeg := NewTIntegerVectorPtr(nbVar);
			pgPos := NewTIntegerVectorPtr(nbVar);
			milieu := NewTIntegerVectorPtr(nbVar);
			egNeg := NewTIntegerVectorPtr(nbVar);
			egPos := NewTIntegerVectorPtr(nbVar);
			probB := NewTRealVectorPtr(nbVar);
			probR := 0;
			egr := 0;
			pgr := 0;
		end; { with }
	end; { CreatePermuteVect }

	procedure ClearPermuteVect (nombreDeVariables: TInteger;
									var result: TPermuteVect);
		var
			i: TInteger;
	begin
		for i := 1 to nombreDeVariables do
			with result do begin
				ppNeg^[i] := 0;
				pgPos^[i] := 0;
				milieu^[i] := 0;
				egNeg^[i] := 0;
				egPos^[i] := 0;
				probB^[i] := 0;
			end; { with }
		with result do begin
			nbVar := nombreDeVariables;
			probR := 0;
			egr := 0;
			pgr := 0;
		end; { with }
	end; { ClearPermuteVect }

	procedure DisposePermuteVect (var result: TPermuteVect);
		procedure Erase (var p: univ Ptr);
		begin
			if p <> nil then begin
				DisposePtr(p);
				p := nil;
			end;
		end; { Erase }

	begin
		with result do begin
			nbVar := 0;
			Erase(ppNeg);
			Erase(pgPos);
			Erase(milieu);
			Erase(egNeg);
			Erase(egPos);
			Erase(probB);
			probR := 0;
			egr := 0;
			pgr := 0;
		end; { with }
	end;

{ Procdure tire de Numerical Recipes, mod. PhC pour changer les positions de indArr }
{ en mme temps que randArr }
	procedure ShellSort (n: integer;
									var randArr: TRealVectorPtr; { le vecteur  trier }
									var indepArr: univ ptr; { vecteur d'entiers ou de rels  entraner }
									isInteger: Boolean);
		label
			3;
		const
			aln2i = 1.442695022;
			tiny = 1.0e-5;
		var
			nn, m, lognb2, l, k, j, i: integer;
			t: real;
			ti: Integer;
			intArr: TIntegerVectorPtr;
			realArr: TRealVectorPtr;
	begin
{ Modification PhC 07/05/98 pour entraner un vecteur d'entiers ou de rels }
		if isInteger then
			intArr := TIntegerVectorPtr(indepArr)
		else
			realArr := TRealVectorPtr(indepArr);

{$IFC not undefined THINK_PASCAL}
		lognb2 := trunc(ln(n) * aln2i + tiny);
{$ELSEC}
		lognb2 := system.trunc(ln(n) * aln2i + tiny);
{$ENDC}
		m := n;
		for nn := 1 to lognb2 do begin
			m := m div 2;
			k := n - m;
			for j := 1 to k do begin
				i := j;
3:
				l := i + m;
				if (randArr^[l] < randArr^[i]) then begin
					t := randArr^[i];
					randArr^[i] := randArr^[l];
					randArr^[l] := t;
					if isInteger then begin { Mod. pour entraner un vect d'entiers ou de rels }
						ti := intArr^[i];
						intArr^[i] := intArr^[l];
						intArr^[l] := ti;
					end
					else begin
						t := realArr^[i];
						realArr^[i] := realArr^[l];
						realArr^[l] := t;
					end;
					i := i - m;
					if (i >= 1) then
						goto 3;
				end
			end
		end
	end; { ShellSort }

	function CompteNbGroupes (group: TIntegerVectorPtr;
									n: TInteger): TInteger;
		var
			i, max: TInteger;
	begin
		max := 0;
		for i := 1 to n do
			if group^[n] > max then
				max := group^[n];
		CompteNbGroupes := max;
	end; { CompteNbGroupes }

	function NombreElementsGroupe (group: TIntegerVectorPtr;
									numeroGroupe, n: TInteger): TInteger;
		var
			i, max: TInteger;
	begin
		max := 0;
		for i := 1 to n do
			if group^[i] = numeroGroupe then
				max := max + 1;
		NombreElementsGroupe := max;
	end; { NombreElementsGroupe }

	procedure PermuterUnVecteur (var v: TMatrixPtr;
									original: TMatrixPtr;
									permutationsLimitees: Boolean);
		var
			aleatoireVect: TRealVectorPtr;
			ind: TintegerVectorPtr; { vecteur d'indirection local }
			i, j, k, n, nbElemGroup: TInteger;


{ Cette procdure remplit aleatoireVect de... nombres alatoires }
		procedure RandomisationVecteurAlea (max: TInteger);
			var
				ii: TInteger;
		begin
			for ii := 1 to max do { remplir le vecteur avec des nombres alatoires }
				aleatoireVect^[ii] := Ran3;
{				aleatoireVect^[ii] := Random; // Fonction QuickDraw }
		end; { RandomisationVecteurAlea }

{ Cette procdure effectue la deuxime permutation dans la mthode des doubles }
{ permutations, c'est--dire la permutation de la topologie de l'arbre grce  }
{ la permutation de la sur-diagonale. La permutation des feuilles de l'arbre }
{ est assure par la permutation " la Mantel" effectue par ElementPermute, }
{ plus haut. }
		procedure DoublePerm (var nouveauUltra: TMatrixPtr;
										originalUltra: TMatrixPtr);
			var
				ii, jj, iijjMoins1, jjMoins1jj, iijj,{}
				indicePermuteSurDiagonale, indiceCourantSurDiagonale: TInteger;
		begin
{ Notez le "n-1" au lieu du "n" attendu, puisque la sur-diagonale n'a que }
{ n-1 lments, et non pas n }

			RandomisationVecteurAlea(n);
{ Mettre dans le vecteur d'indirection locale les indices de la sur-diagonale }
			for ii := 1 to n - 1 do
{ Quel est l'lment dans mon vecteur dpli "sur" la sur-diagonale? }
				ind^[ii] := IJVersElement(n, ii, ii + 1); { Note: n et non pas n-1 ici }

{ Trier le vecteur alatoire, et celui d'indirection en mme temps }
			ShellSort(n - 1, aleatoireVect, ind, INTEGER_ARRAY);

{ On a maintenant de quoi permuter la sur-diagonale de la matrice ultramtrique }
{ originale (reprsente sour forme de vecteur). }
{ Permutation de cette sur-diagonale s'effectue ainsi }
{ (cf. Lapointe & Legendre, J. Class. 8:177-200 (1991). }
			for ii := 1 to n - 1 do begin
{ Quel est l'lment dans mon vecteur dpli "sur" la sur-diagonale? }
				indicePermuteSurDiagonale := ind^[ii];
				indiceCourantSurDiagonale := IJVersElement(n, ii, ii + 1);
				SetElement(nouveauUltra, indicePermuteSurDiagonale, 1, GetElement(originalUltra, indiceCourantSurDiagonale, 1));
			end; { for ii }

{ La sur-diagonale de nouveauUltra est maintenant remplie des valeurs permutes, il ne reste qu' }
{ reconstruire une topologie alatoire  l'aide de l'ingalit triangulaire }
			for ii := 1 to n - 2 do
				for jj := ii + 2 to n do begin
					iijjMoins1 := IJVersElement(n, ii, jj - 1);
					jjMoins1jj := IJVersElement(n, jj - 1, jj);
					iijj := IJVersElement(n, ii, jj);
					if GetElement(nouveauUltra, iijjMoins1, 1) > GetElement(nouveauUltra, jjMoins1jj, 1) then
						SetElement(nouveauUltra, iijj, 1, GetElement(nouveauUltra, iijjMoins1, 1))
					else
						SetElement(nouveauUltra, iijj, 1, GetElement(nouveauUltra, jjMoins1jj, 1));
				end; { for jj }

{ La permutation de la topologie est maintenant termine! }

		end; { DoublePerm }

{ Cette procdure permute les longeurs de branches d'un arbre additif ainsi que la }
{ topologie de l'arbre ultramtrique sous-jacent en extrayant cet arbre et utilisant }
{ DoublePerm. }
		procedure TriplePerm (var additif: TMatrixPtr);
			var
				ultra, ultraPermutee: TMatrixPtr; { Matrice ultramtrique et sa valeur permutee }
				etoileMat, etoileVect: TMatrixPtr; { Arbre toile sous forme de matrice et de vecteur }
				d: TReal;
				i: TInteger;

{ Cette fonction calcule la distance entre deux objets i et j en la lisant dans la }
{ matrice m, qui est une matrice de distance triangulaire suprieure sans diagonale }
{ dplie }
			function Distance (m: TMatrixPtr;
											i, j: TInteger): TReal;
				var
					temp: TInteger;
			begin
				if i > j then begin
{ changer les valeurs de i et j pour se retrouver sur la triangulaire suprieure }
					temp := i;
					i := j;
					j := temp;
				end; { if }

				if i = j then { facile! }
					Distance := 0
				else begin
					temp := IJVersElement(m^.nbObjets, i, j); { Trouver l'lment en question }
					Distance := GetElement(m, temp, 1);
				end; { else }

			end; { Distance }

			procedure RemplirEtoileMat;
				var
					i, j, e: TInteger;
					dI, dJ: TReal;
			begin
				for i := 1 to additif^.nbObjets - 1 do
					for j := i + 1 to additif^.nbObjets do begin
{ e est l'lment i,j de la triangulaire suprieure }
						e := IJVersElement(additif^.nbObjets, i, j);
{ dI et dJ sont les distances dans le vecteur etoileVect }
{ Note: ElementPermute est l pour la reconstruction lors de la permutation }
						dI := GetElement(etoileVect, ElementPermute(i, etoileVect), 1);
						dJ := GetElement(etoileVect, ElementPermute(j, etoileVect), 1);
{ Additionnons dI et dJ pour connatre la valeur de la distance entre l'lment i et }
{ l'lment j sur l'arbre toile }
						SetElement(etoileMat, e, 1, dI + dJ);
					end; { for j }
			end; { RemplirEtoileMat }

		begin
			ultra := nil;
			ultraPermutee := nil;
			etoileMat := nil;
			etoileVect := nil;
{ Premire tape: extraire l'arbre toile sous forme de vecteur de la matrice additive }
{ Ce vecteur reprsente la distance de la racine  la feuille concerne. Il est donc }
{ aussi long qu'il y a de feuilles (nbObjets) }
			if not NewMatrix(etoileVect, additif^.nbObjets, 1, false, false, permVector) then
				; { Ce vecteur sera permut de faon simple }
			CreerVecteurIndirection(etoileVect);

			for i := 1 to additif^.nbObjets do begin
				d := Distance(additif, additif^.root, i);
				SetElement(etoileVect, i, 1, d);
			end; { for i }

{ Voil, dans etoileVect on a maintenant les distances des branches de l'arbre-toile. }
{ Construisons la matrice de distance qui y correspond; toujours une triangulaire }
{ suprieure dplie: }
{ D'abord, crer cette matrice. Elle a les mmes caractristiques qu'additif }
{ Note: cette matrice ne sera pas permute }
			if not NewMatrix(etoileMat, additif^.n, additif^.p, false, false, permNone) then
				;
{ Ensuite, la remplir avec le contenu de etoileVect }
			RemplirEtoileMat;

{ Cration de la matrice ultramtrique, encore une fois identique  additif }
			if not NewMatrix(ultra, additif^.n, additif^.p, false, false, permNone) then
				;

{ Soustrayons l'arbre-toile, en forme de matrice,  additif pour obtenir ultra }
			for i := 1 to additif^.n do begin
{ d: diffrence entre arbre additif et toile }
				d := GetElement(additif, i, 1) - GetElement(etoileMat, i, 1);
				SetElement(ultra, i, 1, d);
			end; { for i }

{ Voil! On a un arbre ultramtrique dans ultra! }
{ Copions-le dans "ultraPermutee" }
			if not NewMatrix(ultraPermutee, ultra^.n, ultra^.p, false, false, permDouble) then
				; { Ce vecteur sera permut en double permutation }
			CreerVecteurIndirection(ultraPermutee);
{ et permutons-le tout-de-suite... }
			DoublePerm(ultraPermutee, ultra);

{ Permutons l'arbre-toile }
			PermuterUnVecteur(etoileVect, nil, false);
{ Reconstruire l'arbre-toile en forme de matrice avec l'arbre-toile permut }
			RemplirEtoileMat; { Marche encore  cause de "ElementPermute" }

{ Additionner ultraPermutee et etoileMat pour reconstituer l'arbre additif permut }
			for i := 1 to additif^.n do begin
{ d: somme de ultraPermutee et toile }
				d := GetElement(ultraPermutee, i, 1) + GetElement(etoileMat, i, 1);
				SetElement(additif, i, 1, d);
			end; { for i }

{ Et voil!  ce niveau on a un arbre additif permut... }

			DisposeMatrix(ultra);
			DisposeMatrix(ultraPermutee);
			DisposeMatrix(etoileMat);
			DisposeMatrix(etoileVect);
		end; { TriplePerm }

	begin
		Assert(v <> nil);

		if v^.permMeth <> permNone then begin { On dsire permuter }
			ind := nil;

			if v^.permMeth = permVector then { Taille du vecteur de nombres alatoires }
				n := v^.n
			else
				n := v^.nbObjets;
			aleatoireVect := NewTRealVectorPtr(n);

			if not permutationsLimitees then begin
{ On ne tient pas compte des groupes ventuels dans la matrice }
{ C'est la mthode normale de permutation }
				RandomisationVecteurAlea(n);
{ Trier ce vecteur, et celui d'indirection en mme temps }
				ShellSort(n, aleatoireVect, v^.indirection, INTEGER_ARRAY);
			end
			else begin { Permutations limites par les groupes de la matrice }
				Assert(v^.group <> nil);

{ Crer le vecteur d'indirection locale qui servira aux permutations  l'intrieur d'un groupe }
				ind := NewTIntegerVectorPtr(n); { n = taille max, je suis paresseux }
				if (ind = nil) then
					ProblemAlert(errNoMatrixMemory, 0, 'ind:PermuterUnVecteur:Limit', '');

				for i := 1 to CompteNbGroupes(v^.group, n) do begin { permutation pour chaque groupe }
					nbElemGroup := NombreElementsGroupe(v^.group, i, n);

{ initialiser le vecteur d'indirection avec les vraies valeurs du vecteur v^.indirection }
					k := 1;
					for j := 1 to n do
						if v^.group^[j] = i then begin {ʎlment du bon groupe }
							ind^[k] := j;
							k := k + 1;
						end;

					RandomisationVecteurAlea(nbElemGroup);
					ShellSort(nbElemGroup, aleatoireVect, ind, INTEGER_ARRAY);

{ Maintenant, ind contient les indices permuts des nbElemGroup lments du groupe i }
{ il suffit de replacer les bonnes valeurs dans le vecteur "v^.indirection" }
					k := 1;
					for j := 1 to n do
						if v^.group^[j] = i then begin {ʎlment du bon groupe }
							v^.indirection^[j] := ind^[k];
							k := k + 1;
						end;

				end; { for i }

				DisposePtr(Ptr(ind));
				ind := nil; { paranod }

			end; { permutations limites }

{ Voil, le vecteur d'indirections a t permut... }

{ Maintenant, faire la permutation double ou triple si demand }
			if v^.permMeth in [permDouble, permTriple] then begin
				Assert(original <> nil);
				if (original = nil) then
					MyDebugStr('Must pass a non-nil original vector for double or triple permutation');

{ Crer le vecteur d'indirection locale qui servira aux doubles et triples permutations }
				ind := NewTIntegerVectorPtr(n);
				if (ind = nil) then
					ProblemAlert(errNoMatrixMemory, 0, 'ind:PermuterUnVecteur', '');
				for i := 1 to n do
					ind^[i] := i;

				if v^.permMeth = permDouble then
					DoublePerm(v, original)
				else
					TriplePerm(v);

				DisposePtr(Ptr(ind));
				ind := nil; { paranod }
			end; {if  permDouble, permTriple }

			DisposePtr(Ptr(aleatoireVect));

		end { if permMeth <> permNone }
		else
			MyDebugStr('PermuterUnVecteur: permMeth = permNone');

	end; { PermuterUnVecteur }

	procedure InitRan3 (initSeed, initCount: integer);
		var
			i: integer;
			zzz: real;
	begin
		gSeed := initSeed;
		for i := 1 to initCount do
			zzz := Ran3;
{			zzz := Random;}
	end; { InitRan3 }

{ Note sur la gnration de nombres alatoires                                }
{ PhC a investigu divers gnrateurs de nombres alatoires dont:             }
{ - GGUBFS (voir autres fichiers)                                             }
{ - Random (gnrateur d'entiers alatoire du Systme Macintosh               }
{ - Ran3 (ci-dessous) de Numerical Recipes.                                   }
{ Pour une distribution la plus uniforme possible, je recommande Ran3         }
{ et c'est ce que j'utilise ici. Ran3 est le plus uniforme des 3 lors d'un    }
{ test de chi-carr sur une distribution de 5000 nombres alatoires avec une  }
{ probabilit de 4%. Random suit  1% et GGUBFS est loin derrire avec 0.01%. }
{ videmment, une probabilit leve dnote l'adhsion  l'hypothse nulle    }
{ d'uniformit de la distribution...                                          }
{ Modification PhC 18/09/98: j'ai enlev le paramtre idum de la fonction     }
{ Ran3, remplac par gSeed (qui est initialis  l'appel de InitRan3). Cela   }
{ devrait simplifier les appels  Ran3 (pas besoin de dclarer une variable   }
{ 'seed' locale).                                                             }

{ Procdure de Numerical Recipes }
	function Ran3: real;
		const
			mbig = 4.0e6;
			mseed = 1618033.0;
			mz = 0.0;
			fac = 2.5e-7;
		var
			i, ii, k: integer;
			mj, mk: real;
	begin
		if (gSeed < 0) then begin
			mj := mseed + gSeed;
			if mj >= 0.0 then
{$IFC not undefined THINK_PASCAL}
				mj := mj - mbig * trunc(mj / mbig)
{$elsec}
				mj := mj - mbig * system.trunc(mj / mbig)
{$endc}
			else
{$IFC not undefined THINK_PASCAL}
				mj := mbig - abs(mj) + mbig * trunc(abs(mj) / mbig);
{$elsec}
			mj := mbig - abs(mj) + mbig * system.trunc(abs(mj) / mbig);
{$endc}
			glma[55] := mj;
			mk := 1;
			for i := 1 to 54 do begin
				ii := 21 * i mod 55;
				glma[ii] := mk;
				mk := mj - mk;
				if (mk < mz) then
					mk := mk + mbig;
				mj := glma[ii]
			end;
			for k := 1 to 4 do begin
				for i := 1 to 55 do begin
					glma[i] := glma[i] - glma[1 + ((i + 30) mod 55)];
					if (glma[i] < mz) then
						glma[i] := glma[i] + mbig
				end
			end;
			glinext := 0;
			glinextp := 31;
			gSeed := 1
		end;
		glinext := glinext + 1;
		if (glinext = 56) then
			glinext := 1;
		glinextp := glinextp + 1;
		if (glinextp = 56) then
			glinextp := 1;
		mj := glma[glinext] - glma[glinextp];
		if (mj < mz) then
			mj := mj + mbig;
		glma[glinext] := mj;
		Ran3 := mj * fac
	end; { Ran3 }

end. { unit Permutations }