unit PermutePrefs;

{ Deals with preference files }

interface

	uses
{$ifc undefined THINK_PASCAL}
		Types, Resources, QuickDraw, Menus, Dialogs, Fonts, ToolUtils;
{$elsec}
		InterfacesUI;
{$endc}

	const
{ PREF resource info }
		kPrefOSType = 'PREF';
		kPrefDefaultID = 128;
		kPrefNewID = 0;
{ Dialogs }
		kPrefsDlogID = 140;
		iFontPopup = 3;
		iSizePopup = 4;
		iDecimalPopup = 5;
{ Menus }
		mFont = 60;
		mSize = 61;
		mDec = 62;

	type
		TPrefHandle = ^TPrefPtr;
		TPrefPtr = ^TPref;
		TPref = record
				DefaultFont, DefaultSize, DecimalsInResult: Integer;
			end; { TPref }

	function ChangePreferences: Boolean;
	procedure GetFontAndSize (var font, size: Integer);
	function GetDecimals: Integer;

implementation

{ Returns a handle to the prefs if writeOut is FALSE, writes the prefs if writeOut is TRUE }
{ Returns an error code }
	function ReadWritePreferences (var h: Handle;
									writeOut: Boolean): OSErr;
		var
			err: OSErr;
			dummyH: Handle;
	begin
		err := noErr;
		if writeOut then begin { write out the resource }
				{ Erase whatever resource might be there already }
			dummyH := Get1Resource(kPrefOSType, kPrefNewID);
			if dummyH <> nil then
				RemoveResource(dummyH);
			AddResource(h, kPrefOSType, kPrefNewID, '');
			err := ResError;
			UpdateResFile(CurResFile);
		end
		else begin { read in the resource }
			h := Get1Resource(kPrefOSType, kPrefNewID);
			err := ResError;
			if (h = nil) then begin
				h := Get1Resource(kPrefOSType, kPrefDefaultID); { get default if there is no new preferences }
				err := ResError;
			end;
{if (h <> nil) then begin}
{DetachResource(h);	 Keep in memory }
{end;}
		end; { else }
		ReadWritePreferences := err;
	end; { ReadWritePreferences }

	function ChangePreferences: Boolean;
		var
			oldPort: GrafPtr;
			prefH: TPrefHandle;
			dlg: DialogPtr;
			err: OSErr;
			iType: Integer;
			iHandle: Handle;
			iRect: Rect;
			font, size, nDec, itemHit: Integer;
			menu: MenuHandle;
			fontName, sizeStr: Str255;

		procedure InitialValues;
			var
				i: Integer;
				s: Str255;
		begin
		{ Default num decimals }
			nDec := prefH^^.DecimalsInResult;
			GetDialogItem(dlg, iDecimalPopup, iType, iHandle, iRect);
			i := GetControlMaximum(ControlHandle(iHandle));
			if nDec > (i - 1) then { -1 b/c starts at 0 decimals and menu starts at 1 }
				nDec := i - 1;
			if nDec < 0 then
				nDec := 0;
			SetControlValue(ControlHandle(iHandle), nDec + 1);
		{ Default font and size }
			GetDialogItem(dlg, iFontPopup, iType, iHandle, iRect);
			font := prefH^^.DefaultFont;
			menu := GetMenu(mFont);
			GetFontName(font, fontName);
			iType := 1; { default value }
			for i := 1 to CountMItems(menu) do begin
			{ Clear the menu of check marks }
				CheckItem(menu, i, FALSE);
				GetMenuItemText(menu, i, s);
				if s = fontName then
					iType := i; { found the font }
			end; { for i }
			SetControlValue(ControlHandle(iHandle), iType);

			GetDialogItem(dlg, iSizePopup, iType, iHandle, iRect);
			size := prefH^^.DefaultSize;
			menu := GetMenu(mSize);
{			NumToString(size, sizeStr); }
			sizeStr := StringOf(size);
			iType := 1; { default value }
			for i := 1 to CountMItems(menu) do begin
				GetMenuItemText(menu, i, s);
			{ Clear the menu of check marks }
				CheckItem(menu, i, FALSE);
				if s = sizeStr then
					iType := i; { found the size }
			end; { for i }
			SetControlValue(ControlHandle(iHandle), iType);
		end; { InitialValues }

		procedure GetValues;
			var
				i: Integer;
				dumLongInt: LongInt;
		begin
			GetDialogItem(dlg, iDecimalPopup, iType, iHandle, iRect);
			prefH^^.DecimalsInResult := GetControlValue(ControlHandle(iHandle)) - 1;

			GetDialogItem(dlg, iFontPopup, iType, iHandle, iRect);
			menu := GetMenu(mFont);
			GetMenuItemText(menu, GetControlValue(ControlHandle(iHandle)), fontName);
			GetFNum(fontName, prefH^^.DefaultFont);

			GetDialogItem(dlg, iSizePopup, iType, iHandle, iRect);
			menu := GetMenu(mSize);
			i := GetControlValue(ControlHandle(iHandle));
			GetMenuItemText(menu, i, sizeStr);
{			StringToNum(sizeStr, dumLongInt);}
			ReadString(sizeStr, dumLongInt);
			prefH^^.DefaultSize := LoWord(dumLongInt);
		end; { GetValues }

	begin
		itemHit := cancel; { valeur initiale }
		GetPort(oldPort);
		err := ReadWritePreferences(Handle(prefH), FALSE);
		if (err = noErr) then begin { We have some preferences }
			dlg := GetNewDialog(kPrefsDlogID, nil, DialogPtr(-1));
			err := SetDialogDefaultItem(dlg, OK);
			err := SetDialogCancelItem(dlg, cancel);

		{ Set up initial values }
			InitialValues;

			ShowWindow(dlg);
			SetPort(dlg);
			repeat
				ModalDialog(nil, itemHit);
			until itemHit in [ok, cancel];

			if itemHit = OK then begin
				GetValues;
				err := ReadWritePreferences(Handle(prefH), TRUE);
			end; { if OK }

			DisposeDialog(dlg);
{DisposeHandle(Handle(prefH));}
		end
		else
			; { no prefs resource! }

		SetPort(oldPort);
		ChangePreferences := (itemHit = ok);
	end; { ChangePreferences }

	procedure GetFontAndSize (var font, size: Integer);
		var
			h: TPrefHandle;
			err: OSErr;
	begin
		font := 0;
		size := 12;
		err := ReadWritePreferences(Handle(h), false);
		if (h <> nil) and (err = noErr) then begin
			font := h^^.DefaultFont;
			size := h^^.DefaultSize;
		end;
	end; { GetFontAndSize }

	function GetDecimals: Integer;
		var
			h: TPrefHandle;
			err: OSErr;
	begin
		err := ReadWritePreferences(Handle(h), false);
		if (h <> nil) and (err = noErr) then
			GetDecimals := h^^.DecimalsInResult
		else
			GetDecimals := 3;
	end; { GetDecimals }

end. { PermutePrefs }