unit PermuteProgressInfo;

interface

	uses
{$ifc undefined THINK_PASCAL}
		Types, Windows, Quickdraw, 
{$endc}
		MathDeclarations, Assertions, TransDisplay;

	const
		kProgressBannerHeight = 15; { Height in pixel of the small banner at the bottom of display windows }
		kProgressOffset = 3; { Offset of progress rectangle }
		kProgressLength = 100; { Length of progress rectangle }

	type
		ProgressInfoRecord = record
				displayWindow: WindowPtr;
				width: Integer;
				nbCompForOnePixel: Real;
				useThreads: Boolean;
			end; { record }
		ProgressInfoPtr = ^ProgressInfoRecord;

	procedure CalculateProgressRect (var r: Rect;
									port: GrafPtr);

	procedure InitProgress (min, max: TInteger;
									progress: ProgressInfoPtr);

	procedure ProgressInfo (count: TInteger;
									progress: ProgressInfoPtr);

implementation

	procedure CalculateProgressRect (var r: Rect;
									port: GrafPtr);
	begin
		r := port^.portRect;
		r.bottom := r.top + kProgressBannerHeight;
		r.right := r.right - 15; { account for scroll bar }
		InsetRect(r, kProgressOffset, kProgressOffset);
		r.right := r.left + kProgressLength;
	end; { CalculateProgressRect }
{}
{			with display^.portRect do begin}
{				tir.progress^.progressRect.left := left + kProgressOffset;}
{				tir.progress^.progressRect.top := bottom - kProgressBannerHeight + kProgressOffset;}
{				tir.progress^.progressRect.right := left + kProgressOffset + kProgressLength;}
{				tir.progress^.progressRect.bottom := bottom - kProgressOffset;}
{			end; with }


	procedure InitProgress (min, max: TInteger;
									progress: ProgressInfoPtr);
		var
			total: LongInt;
			oldPort: GrafPtr;
			progressRect: Rect;
	begin
		Assert(progress <> nil);
		total := Abs(max - min); { force it to be a longint for division below }
		CalculateProgressRect(progressRect, progress^.displayWindow);
		with progress^ do begin
			width := progressRect.right - progressRect.left;
			nbCompForOnePixel := total / width;
		end; { with }

		GetPort(oldPort);
		SetPort(progress^.displayWindow);

		CalculateProgressRect(progressRect, progress^.displayWindow);

		EraseRect(progressRect);

		SetPort(oldPort);
	end; { InitProgress }

	procedure ProgressInfo (count: TInteger;
									progress: ProgressInfoPtr);
		var
			oldPort: GrafPtr;
			rw: Integer;
			r: Rect;
			p: Pattern;
	begin
		Assert(progress <> nil);
		CalculateProgressRect(r, progress^.displayWindow);
{$ifc undefined THINK_PASCAL}
		rw := system.trunc((count / progress^.nbCompForOnePixel) + 0.5);
		p := qd.black;
{$elsec}
		rw := trunc(count * progress^.nbCompForOnePixel);
		p := black;
{$endc}
		GetPort(oldPort);
		SetPort(progress^.displayWindow);

		r.right := r.left + rw;
		FillRect(r, p);

		SetDWindow(oldPort);
	end; { ProgressInfo }

end. { unit PermuteProgressInfo }