unit Statistiques;

interface

	uses
		MathDeclarations, MatrixAlgebra, Permutations, Assertions;

	const
{ Pour rendre les appels  "RegressionMultiple" plus lisibles }
		Pas_Besoin_Centrer_Reduire = False;
		Besoin_Centrer_Reduire = True;

	procedure CentrerEtReduire (var m: TMatrixPtr;
									c: TInteger);

{ Fonction calculant [X'X]^-1  partir d'un X quelconque, utile en rgression }
	function CreerXPrimeXMoins1 (x: TMatrixPtr): TMatrixPtr;

{ Fonction qui trouve le vecteur b rsultat de [X'X]^-1 X'Y }
{ ainsi que les valeurs de t associes  ces b }
{ On peut fournir la matrice xPrimeXMoins1 si on dsire aller plus vite }
{ (i.e. espace dj allou) , sinon fournir nil. }
{ On peut aussi fournir un vecteur d'indirection  utiliser lors de tests par permutations }
{ (si pas de test par permutations, fournir nil); dans ce cas signifier si les permutations }
{ sont en mode mantel (matrice de distance) ou non }
	procedure RegressionMultiple (var y, x, b, t: TMatrixPtr;
									var rCarre: TReal;
									xPrimeXMoins1: TMatrixPtr;
									besoinCentrerReduireXetY: Boolean);

{	Module:       f.c}
{	Purpose:      compute approximations to F distribution probabilities}
{	Contents:     pof(), critf()}
{	Programmer:   Gary Perlman}
{	Organization: Wang Institute, Tyngsboro, MA 01879}
{	Tester:       compile with -DFTEST to include main program}
{	Copyright:    none}
{	Tabstops:     4}
{ Pascal translation by Philippe Casgrain }
	function PoF (F: TReal;
									df1, df2: integer): TReal;
	function CritF (p: TReal;
									df1, df2: Integer): TReal;

{ Cette procdure remplace les valeurs d'un vecteur-colonne par les rangs correspondants }
{ incluant les corrections pour rangs gaux }
	procedure TransformerEnRang (var orig: TMatrixPtr);

{ Calcul du Tau de Kendall entre ces deux vecteurs-colonnes. Suppose que ces vecteurs }
{ sont dj transforms en rangs (voir ci-haut). }
	function CalcTauKendall (matA, matB: TMatrixPtr): TReal;

{ Fonction A. Vaudor, mod. PhC 04/06/98 Extended->TReal, tire de Statis.p }
	function Gauss_Vaudor (X: TReal): TReal;

implementation

	const
		I_PI = 0.3183098861837906715377675;
		F_EPSILON = 0.000001;       (* accuracy of critf approximation *)
		F_MAX = 9999.0;           (* maximum F ratio *)

{ Centrer et rduire la colonne c d'une matrice m }
	procedure CentrerEtReduire (var m: TMatrixPtr;
									c: TInteger);
		var
			moyenne, ecartType, somme, sommeCarres, val: TReal;
			i: TInteger;
	begin
		Assert(m <> nil);
		if (c > 0) and (c <= m^.p) then begin
			somme := 0;
			sommeCarres := 0;
			for i := 1 to m^.n do begin
				val := GetElement(m, i, c);
				somme := somme + val;
				sommeCarres := sommeCarres + (val * val);
			end; { for i }

			moyenne := somme / m^.n;
			ecartType := Sqrt((sommeCarres - (somme * somme) / m^.n) / (m^.n - 1));

			for i := 1 to m^.n do begin
				val := GetElement(m, i, c);
				val := (val - moyenne) / ecartType;
				SetElement(m, i, c, val);
			end; { for i }
		end;
	end; { CentrerEtReduire }

{ m: moyenne }
{ sx: cart-type }
{ n: nombre d'lments }
	function tDeStudent (m1, m2, sx1, sx2: TReal;
									n1, n2: TInteger): TReal;
	begin
{  arranger plus tard... }
		tDeStudent := (m1 - m2) / (sx1 / n1 + sx2 / n2);
	end; { tDeStudent }

{ Cette fonction rsout le problme suivant:                           }
{ Si on reprsente une matrice de distance par un long vecteur dpli, }
{ et que ce vecteur n'est que la triangulaire suprieure de la matrice }
{ de distance originale, et si cette matrice de distance a n lignes et }
{ n colonnes, la longueyr x du vecteur est (n*(n-1))/2.                }
{ Ca, c'est facile. Mais si on a x, que vaut n? C'est bien sr une     }
{ quation du second degr:                                            }
{ n*(n-1)/2 = x <=> n^2 - n - 2x = 0                                   }
{ Dont la solution (positive) est:                                     }
{ n = (1 + (8x+1))/2                                                  }
{ Si cette solution est entire, alors on peut dire qu'il y a juste    }
{ assez d'lments x dans le vecteur considr pour reprsenter n      }
{ objets. Sinon, on retourne 0.                                        }

	function CreerXPrimeXMoins1 (x: TMatrixPtr): TMatrixPtr;
		var
			xPrimeXMoins1: TMatrixPtr;
{$ifc false}
			i, j: integer;
{$endc}
	begin
		xPrimeXMoins1 := nil;
		MMult(x, x, xPrimeXMoins1, true); { Pas encore invers, voir ci-dessous }
		Assert(xPrimeXMoins1 <> nil);
		if not MInverse(xPrimeXMoins1) then
			MyDebugStr('Error: X''X matrix singular (can''t be inverted)');
		CreerXPrimeXMoins1 := xPrimeXMoins1;
{$ifc do_debugfiles}
		writeln(debug1, 'X''X-1');
		for i := 1 to 6 do begin
			for j := 1 to 6 do
				write(debug1, GetElement(xPrimeXMoins1, i, j) : 8 : 6, '   ');
			writeln(debug1);
		end;
{$endc}
	end; { CreerXPrimeXMoins1 }

	procedure RegressionMultiple (var y, x, b, t: TMatrixPtr;
									var rCarre: TReal;
									xPrimeXMoins1: TMatrixPtr;
									besoinCentrerReduireXetY: Boolean);
		var
			i, j: TInteger;
			disposeXPrimeXMoins1: Boolean;
			xPrimeY, xb, residus: TMatrixPtr;
			ssqRes, r, sigma, seB: TReal;

		function CalculerRCarre: TReal;
			var
				rXY: TMatrixPtr;
				i, j: TInteger;
				sum: TReal;
{Note: yBarre = 0, xBarre[i] = 0 car on a centre et reduit}
{  De meme, les ecarts-types sur y et sur x sont egaux a 1}
		begin
			sum := 0;
			rXY := nil;
			DuplicateMatrix(b, rXY, 0, 0, false); { Crer le vecteur en le copiant sur b }
			for j := 1 to x^.p do begin
				sum := 0;
				for i := 1 to x^.n do begin { ou y^.n, c'est la mme chose! }
					sum := sum + (GetElement(y, ElementPermute(i, y), 1) * GetElement(x, i, j));
				end;
				SetElement(rXY, j, 1, sum / (x^.n - 1));
			end; {for j}
			sum := 0;
			for i := 1 to b^.n do begin
				sum := sum + (GetElement(b, i, 1) * GetElement(rXY, i, 1));
			end; {for j}
			DisposeMatrix(rXY);
			CalculerRCarre := sum;
		end; { function CalculerRCarre }

	begin
		Assert(y <> nil);
		Assert(x <> nil);
		Assert(x^.n > 0);
		Assert(x^.p > 0);
		Assert(x^.n > x^.p);

		xPrimeY := nil;
		xb := nil;
		residus := nil;

		if besoinCentrerReduireXetY then begin { Centrer et rduire pour ne pas avoir d'intercept }
			if x <> nil then
				for j := 1 to x^.p do
					CentrerEtReduire(x, j);
{ Il n'y a pas de permutations  faire lors d'un centrage et rduction... }
			CentrerEtReduire(y, 1);
		end; { if besoinCentrerReduireXetY }

{ Vrifier si ces matrices sont fournies ou non, pour savoir }
{ si on doit en disposer nous-mmes ou non. }
		disposeXPrimeXMoins1 := (xPrimeXMoins1 = nil);

{ Calcul du vecteur b }

{ Si l'usager fournit xPrimeXMoins1, on suppose qu'elle a dj la bonne valeur }
		if disposeXPrimeXMoins1 then
			xPrimeXMoins1 := CreerXPrimeXMoins1(x);

		MMult(x, y, xPrimeY, true);
		Assert(xPrimeY <> nil);
		MMult(xPrimeXMoins1, xPrimeY, b, false); {voici le vecteur b! }

{ Calcul du r carre }
		rCarre := CalculerRCarre;

{ Calculer le vecteur de pseudo-t s a partir du vecteur des b s }
		DuplicateMatrix(b, t, 0, 0, false); { Crer le vecteur en le copiant sur b }
		Assert(t <> nil);
{ D'abord, calculer le vecteur XB, produit de la matrice X et du vecteur B }
		MMult(x, b, xb, false);
		Assert(xb <> nil);
{ Ensuite, calculer le vecteur des rsidus }
		if NewMatrix(residus, xb^.n, xb^.p, false, false, permNone) then
			;
		Assert(residus <> nil);
		for i := 1 to residus^.n do
			SetElement(residus, i, 1, GetElement(y, ElementPermute(i, y), 1) - GetElement(xb, i, 1));
{ Ensuite la somme de carrs des rsidus }
		ssqRes := 0;
		for i := 1 to residus^.n do begin
			r := GetElement(residus, i, 1);
			ssqRes := ssqRes + (r * r);
		end;
{ Here is the estimate of the residual mean square "sigma square hat" }
		sigma := ssqRes / (x^.n - x^.p - 1);
{ Finally the B's standard errors can be computed, and from that the t's }
{ (which are just b's divided by their standard errors }
		for i := 1 to t^.n do begin
			seB := Sqrt(sigma * GetElement(xPrimeXMoins1, i, i));
			if seB = 0 then { viter les divisions par zro si corrlation entre deux variables identiques }
				seB := 1;
			SetElement(t, i, 1, GetElement(b, i, 1) / seB);
		end;

		DisposeMatrix(xPrimeY);
		DisposeMatrix(xb);
		DisposeMatrix(residus);

		if disposeXPrimeXMoins1 then
			DisposeMatrix(xPrimeXMoins1);

	end; { RegressionMultiple }

(*FUNCTION pof: probability of F *)
{ALGORITHM Compute probability of F ratio.}
{	Adapted from Collected Algorithms of the CACM}
{	Algorithm 322}
{	Egon Dorrer}
	function PoF (F: TReal;
									df1, df2: integer): TReal;
		var
			i, j: Integer;
			a, b: Integer;
			w, y, z, d, p: TReal;
	begin

		if ((F < F_EPSILON) or (df1 < 1) or (df2 < 1)) then begin
			pof := 1.0;
			exit(pof);
		end;
		if (df1 mod 2) > 0 then
			a := 1
		else
			a := 2;
		if (df2 mod 2) > 0 then
			b := 1
		else
			b := 2;
		w := (F * df1) / df2;
		z := 1.0 / (1.0 + w);
		if (a = 1) then
			if (b = 1) then begin
				p := sqrt(w);
				y := I_PI; (* 1 / 3.14159 *)
				d := y * z / p;
				p := 2.0 * y * arctan(p);
			end
			else begin
				p := sqrt(w * z);
				d := 0.5 * p * z / w;
			end
		else if (b = 1) then begin
			p := sqrt(z);
			d := 0.5 * z * p;
			p := 1.0 - p;
		end
		else begin
			d := z * z;
			p := w * z;
		end;
		y := 2.0 * w / z;

		j := b + 2;
		while j <= df2 do begin
			d := d * (1.0 + a / (j - 2.0)) * z;
			if a = 1 then
				p := p + d * y / (j - 1.0)
			else
				p := (p + w) * z;
			j := j + 2;
		end;
		y := w * z;
		z := 2.0 / z;
		b := df2 - 2;
		i := a + 2;
		while i <= df1 do begin
			j := i + b;
			d := d * (y * j / (i - 2.0));
			p := p - (z * d / j);
			i := i + 2;
		end;
	(* correction for approximation errors suggested in certification *)
		if (p < 0.0) then
			p := 0.0
		else if (p > 1.0) then
			p := 1.0;
		pof := (1.0 - p);
	end;

(*FUNCTION critf: compute critical F value t produce given probability *)
{ALGORITHM}
{	Begin with upper and lower limits for F values (maxf and minf)}
{	set to extremes.  Choose an f value (fval) between the extremes.}
{	Compute the probability of the f value.  Set minf or maxf, based}
{	on whether the probability is less than or greater than the}
{	desired p.  Continue adjusting the extremes until they are}
{	within F_EPSILON of each other.}
	function CritF (p: TReal;
									df1, df2: Integer): TReal;
		var
			fval: TReal;
			maxf: TReal;     (* maximum F ratio *)
			minf: TReal;       (* minimum F ratio *)
			conv: Integer;
	begin
		maxf := F_MAX;
		minf := 0.0;
		conv := 0;
		if ((p <= 0.0) or (p >= 1.0)) then begin
			critf := (0.0);
			exit(critf);
		end;

		fval := 1.0 / p;             (* the smaller the p, the larger the F *)

		while (abs(maxf - minf) > F_EPSILON) do begin
			if (pof(fval, df1, df2) < p) then     (* F too large *)
				maxf := fval
			else                              (* F too small *)
				minf := fval;
			fval := (maxf + minf) * 0.5;
			conv := conv + 1;
		end;

		critf := fval;
	end;

	procedure TransformerEnRang (var orig: TMatrixPtr);
		var
			copie: TRealVectorPtr;
			indice: TintegerVectorPtr;
			i, ii, n, iIemeElement, ties, sum, nombre, k: TInteger;
			prec, rangMoyen: TReal;
	begin
		Assert(orig <> nil);
		n := orig^.n;
		copie := NewTRealVectorPtr(n);
		indice := NewTIntegerVectorPtr(n);
		if (copie = nil) or (indice = nil) then
			MyDebugStr('Not enough memory TransformerEnRang');

{ premirement, faire une copie de l'original }
		for i := 1 to n do begin
			copie^[i] := GetElement(orig, i, 1); { suppose un seul vecteur pour le moment }
			indice^[i] := i;
		end; { for i }

{ deuximement, trier la copie avec les indices }
		ShellSort(n, copie, indice, INTEGER_ARRAY);

{ Le premier lment du vecteur indice indique maintenant l'lment du vecteur original }
{ qui a le premier rang... Le i-ime lment du vecteur indice indique l'lment du vecteur }
{ original qui a le i-ime rang }

{ troisimement, trouver les ex-aequo et leur attribuer des rangs moyens }
{ Ce faisant, remplacer les valeurs de "copie" par les rangs }
		ties := 0;
		i := 1;
		repeat
			prec := copie^[i];
			ii := i;
			i := i + 1;
			if (i <= n + 1) then  { note: n+1 plutt que n; si i = n+1 alors copie[i] = 0 }

				if (Abs(copie^[i] - prec) < global_epsilon) then begin
					ties := ties + 1;
					repeat
						i := i + 1
					until (i > n) or (copie^[i] - prec >= global_epsilon);
{ Il faut mettre au rang moyen, de ii  i-1 }
					sum := 0;
					nombre := 0;
					for k := ii to i - 1 do begin
						sum := sum + k;
						nombre := nombre + 1;
					end; { for k }
					rangMoyen := sum / nombre;
					for k := ii to i - 1 do
						copie^[k] := rangMoyen;
				end
				else begin { pas de rang gaux, on copie le rang et on continue }
					copie^[ii] := ii;
				end;

		until i > n;

{ finalement, copier les rangs dans le vecteur original }
		for i := 1 to n do begin
			iIemeElement := indice^[i];
			SetElement(orig, iIemeElement, 1, copie^[i]);
		end; { for i }

		DisposePtr(Ptr(copie));
		DisposePtr(Ptr(indice));
	end; { TransformerEnRang }

	function CalcTauKendall (matA, matB: TMatrixPtr): TReal;

{ Cette fonction calcule la somme des ex-aequo, c--d. tx(tx-1). Voir Scherrer pour }
{ de plus amples explications }
{ Suppose que v a dja t tri }
		function CalcSumTies (n: TInteger;
										v: TRealVectorPtr): TInteger;
			var
				i, ii, numTies, sumTies, range: TInteger;
				prec: TReal;
		begin
			numTies := 0;
			sumTies := 0;
			i := 1;
			repeat
				prec := v^[i];
				ii := i;
				i := i + 1;
				if (i <= n + 1) then  { note: n+1 plutt que n; si i = n+1 alors v[i] = 0 }
					if (v^[i] = prec) then begin
						numTies := numTies + 1;
						repeat
							i := i + 1
						until (i > n) or (v^[i] <> prec);
						range := i - ii;
						sumTies := sumTies + (range * (range - 1));
					end

			until i > n;
			CalcSumTies := sumTies;
		end; { CalcSumTies }

		var
			copieA, copieB: TRealVectorPtr;
			i, j, n, pointage, S, sumTx, sumTy: TInteger;
			nNMoins1, n1, n2: TReal;
	begin
		Assert(matA <> nil);
		Assert(matB <> nil);
		Assert(matA^.n = matB^.n);
		n := matA^.n;
{ faire une copie de A et B car on va trier ces vecteurs selon A }
		copieA := NewTRealVectorPtr(n);
		copieB := NewTRealVectorPtr(n);
		if (copieA = nil) or (copieB = nil) then
			MyDebugStr('Not enough memory CalcTauKendall');

{ Remplir les lment de copie A }
		for i := 1 to n do
			copieA^[i] := GetElement(matA, i, 1);
{ Remplir les lments de la copie de B avec les lments permuts de B }
		for i := 1 to n do
			copieB^[i] := GetElement(matB, ElementPermute(i, matB), 1);

{ Trier copieB en entranant copieA, dans le seul but de trouver le nombre d'ex-aequo }
{et de calculer ty(ty-1) }
		ShellSort(n, copieB, copieA, REAL_ARRAY);
		sumTy := CalcSumTies(n, copieB);

{ Trier copieA en entranant copieB }
		ShellSort(n, copieA, copieB, REAL_ARRAY);
{ Calculer tx(tx-1) }
		sumTx := CalcSumTies(n, copieA);

{ Bon! Maintenant le fun commence. Je vais suivre la procdure dcrite par Scherrer (17.7) }
{ pour le calcul du tau de Kendall }
		S := 0;
		for i := 1 to n - 1 do begin
			pointage := 0;
			for j := i + 1 to n do
				if (copieB^[j] > copieB^[i]) and (copieA^[j] > copieA^[i]) then
					pointage := pointage + 1
				else if (copieB^[j] < copieB^[i]) and (copieA^[j] > copieA^[i]) then
					pointage := pointage - 1;
			S := S + pointage;
		end; { for i }

		DisposePtr(Ptr(copieA));
		DisposePtr(Ptr(copieB));

{ Note: l'quation du Tau est dmonte en morceaux pour faciliter le dboguage... }
		n1 := (n * (n - 1)) - sumTx;
		n2 := (n * (n - 1)) - sumTy;
		nNMoins1 := n1 * n2;
		nNMoins1 := Sqrt(nNMoins1);
		CalcTauKendall := (2 * S) / nNMoins1;

	end; { CalcTauKendall }

	function Gauss_Vaudor (X: TReal): TReal;

		var
			Y: TReal;
			Z, Z1: TReal;
			W: TReal;

	begin { Gauss }
					(* REFERENCE: CACM		    *)
					(* ALGORITHME: 209		    *)
					(* AUTEUR:D.IBBETSON		  *)
					(* TITRE:GAUSS			  *)
					(* TRADUIT EN PASCAL PAR A.VAUDOR *)
		if X = 0 then
			Z := 0
		else begin
			Y := ABS(X) / 2;
			if Y >= 3 then
				Z := 1
			else if Y < 1 then begin
				W := Y * Y;
				Z := ((((((((0.000124818987 * W - 0.001075204047) * W + 0.005198775019) * W - 0.019198292004) * W + 0.059054035642) * W - 0.151968751364) * W + 0.319152932694) * W - 0.531923007300) * W + 0.797884560593) * Y * 2
			end
			else begin
				Y := Y - 2;
{ Mod. PhC 11/02/98: coup z en plusieurs morceaux }
				Z1 := (((((((((((-0.000045255659 * Y + 0.000152529290) * Y - 0.000019538132) * Y - 0.000676904986) * Y + 0.001390604284) * Y - 0.000794620820) * Y - 0.002034254874) * Y + 0.006549791214) * Y - 0.010557625006) * Y + 0.011630447319) * Y - 0.009279453341) * Y + 0.005353579108) * Y - 0.002141268741;
				Z := ((Z1) * Y + 0.000535310849) * Y + 0.999936657524
			end
		end;
		if X > 0 then
			Gauss_Vaudor := (Z + 1) / 2
		else
			Gauss_Vaudor := (1 - Z) / 2;
	end;	(* Gauss_Vaudor *)


end.