unit TestsParPermutation;

interface

	uses
		TransDisplay, Assertions, 
{$ifc can_use_threads}
		Threads, 
{$endc}
		MathDeclarations, Permutations, MatrixAlgebra, Statistiques, PermuteProgressInfo;

	const
		IS_STEPWISE = TRUE;

	type
		TTest = (simple, backward, forward, stepwise);

		TestInfoRecord = record
				permMeth: TMethodePermutation;
				root: TInteger;
				nbPerm: TInteger;
				testMeth: TTest;
				pToRemove, pToEnter: TReal;
				mat: TMatrixPtr;
				whichXVariables, whichYVariable: TBoolVectorPtr;
				progress: ProgressInfoPtr;
			end; { record }
		TestInfoPtr = ^TestInfoRecord;


	function FaireTests (ti: TestInfoPtr): LongInt;

implementation


{ Note: ces procdures de test supposent que y et x sont centrs et rduits }

	procedure RegressionMultiplePermutee (nbPerm: TInteger;
									y, x: TMatrixPtr;
									var bOriginal, tOriginal: TMatrixPtr;
									var rCarreOriginal: TReal;
									var permVect: TPermuteVect;
									progress: ProgressInfoPtr);

		var
			xPrimeXMoins1, yPermute, bPermute, tPermute: TMatrixPtr;
			rCarrePermute, valOriginale, valPermutee, valOriginaleVA: TReal;
			i, j: TInteger;

	begin
		xPrimeXMoins1 := nil;
		yPermute := nil;
		bPermute := nil;
		tPermute := nil;

{ Crer [X'X]^-1 pour ce test }
		xPrimeXMoins1 := CreerXPrimeXMoins1(x);

{ Calculer les valeurs originales des statistiques }
		RegressionMultiple(y, x, bOriginal, tOriginal, rCarreOriginal, xPrimeXMoins1, Pas_Besoin_Centrer_Reduire);

		with permVect do begin
			for j := 1 to x^.p do
				if GetElement(tOriginal, j, 1) < 0 then
					egNeg^[j] := egNeg^[j] + 1
				else
					egPos^[j] := egPos^[j] + 1;
			egr := 1;
		end; { with }

{ Copier y dans yPermute, car double et triple sont des mthodes permutationnelles destructrices }
		DuplicateMatrix(y, yPermute, 0, 0, true);

		i := 1;

		InitProgress(0, nbPerm, progress);

		while (i <= nbPerm) do begin { Calcul de la distribution de la statistique permute }

{$ifc can_use_threads}
			if progress^.useThreads then
				if YieldToAnyThread <> noErr then { ignore error }
					;
{$endc}

			ProgressInfo(i, progress);

			PermuterUnVecteur(yPermute, y, false); { Permuter le vecteur Y sans permutations limites }
{ Centrer et rduire au besoin }
			if yPermute^.permMeth in [permDouble, permTriple] then
				CentrerEtReduire(yPermute, 1);

{ Calculer la statistique permute }
			RegressionMultiple(yPermute, x, bPermute, tPermute, rCarrePermute, xPrimeXMoins1, Pas_Besoin_Centrer_Reduire);

			with permVect do
				for j := 1 to x^.p do begin
					valOriginale := GetElement(tOriginal, j, 1);
					valPermutee := GetElement(tPermute, j, 1);
					valOriginaleVA := abs(valOriginale);

{ Nouvelle mthode d'inscription des permutations par PL, A93 }
					if valPermutee < -valOriginaleVA then
						ppNeg^[j] := ppNeg^[j] + 1
					else if abs(valPermutee + valOriginaleVA) < global_epsilon then
						egNeg^[j] := egNeg^[j] + 1
					else if abs(valPermutee - valOriginaleVA) < global_epsilon then
						egPos^[j] := egPos^[j] + 1
					else if valPermutee > valOriginaleVA then
						pgPos^[j] := pgPos^[j] + 1
					else
						milieu^[j] := milieu^[j] + 1;

				end;{ for j }
{  Meme chose, mais pour le r^2 (sauf qu'il est toujours positif)           }
			with permVect do
				if Abs(rCarrePermute - rCarreOriginal) < global_epsilon then
					egr := egr + 1
				else if rCarrePermute > rCarreOriginal then
					pgr := pgr + 1;

			i := i + 1;
		end; { while (i <= nbPerm) }

		with permVect do begin
{ Calculer les probabilites unilaterales associees a chaque tB: }
			for j := 1 to x^.p do
				if GetElement(tOriginal, j, 1) > 0 then { Calcul a droite }
					probB^[j] := (egPos^[j] + pgPos^[j]) / (nbPerm + 1)
				else { Calcul a gauche: Modification PhC 25/04/95 }
					probB^[j] := (egNeg^[j] + ppNeg^[j]) / (nbPerm + 1);
{ Calculer la probabilit du r carr }
			probR := (egr + pgr) / (nbPerm + 1)
		end; { with permVect }

		DisposeMatrix(xPrimeXMoins1);

	end; { RegressionMultiplePermutee }

	procedure AfficherEnTete (nbPerm: TInteger;
									meth: TMethodePermutation;
									ro: TInteger;
									dep: TString;
									e, r: TReal);

	begin
		DisplayLn;
		DisplayStringLn('NOTES: - Equal (=) values include the initial statistic.');
		DisplayStringLn('       - B coefficients are standardized.');
		DisplayStringLn('       - Probabilities of the B coefficients are actually');
		DisplayStringLn('         those of the associated t-statistics.');
		DisplayLn;
		DisplayString('Dependent variable: ');
		DisplayStringLn(dep);
		DisplayString('Permutation method: ');
		case meth of
			permNone: 
				DisplayStringLn('None: should not get this result!');
			permVector: 
				DisplayStringLn('Vector');
			permMatrix: 
				DisplayStringLn('Matrix');
			permDouble: 
				DisplayStringLn('Double');
			permTriple: 
				DisplayStringLn(StringOf('Triple (rooting object: ', ro : 1, ')'));
			otherwise
				DisplayStringLn('Unknown: should not get this result!');
		end; { case }

		DisplayStringLn(StringOf('Number of permutations: ', nbPerm));
		if e > 0 then
			DisplayStringLn(StringOf('p to enter: ', e : 5 : 3));
		if r > 0 then
			DisplayStringLn(StringOf('p to remove: ', r : 5 : 3));
		DisplayLn;
	end; { AfficherEnTete }

	procedure AfficherResultatsComplets (nbPerm: TInteger;
									permVect: TPermuteVect;
									x, bOriginal: TMatrixPtr;
									rCarreOriginal: TReal);
		var
			i, j, nbDecimales: TInteger;
			ts: TString;
	begin
{$ifc undefined THINK_PASCAL}
		nbDecimales := system.Trunc(Ln(nbPerm) / Ln(10)) + 1;
{$elsec}
		nbDecimales := Trunc(Ln(nbPerm) / Ln(10)) + 1;
{$endc}
{  nbDecimales := 3;  valeur par defaut }

		DisplayStringLn('Variable  b coefficient  p<-|o|  p=-|o| -|o|<p<|o|  p=|o|  |o|<p   Unil. prob.');
		DisplayStringLn('----------------------------------------------------------------   -----------');
		for j := 1 to x^.p do begin
			ts := GetColName(x, j);
			DisplayString(ts);
			for i := Length(ts) to 10 do { pad with spaces }
				DisplayString(' ');
			if GetElement(bOriginal, j, 1) > 0 then
				DisplayString(' ');
			DisplayString(StringOf(GetElement(bOriginal, j, 1) : 10 : 8));
			DisplayString('   ');
			DisplayString(StringOf(permVect.ppNeg^[j] : 5));
			DisplayString('   ');
			DisplayString(StringOf(permVect.egNeg^[j] : 5));
			DisplayString('    ');
			DisplayString(StringOf(permVect.milieu^[j] : 5));
			DisplayString('    ');
			DisplayString(StringOf(permVect.egPos^[j] : 5));
			DisplayString('  ');
			DisplayString(StringOf(permVect.pgPos^[j] : 5));
			DisplayString('      ');
			DisplayString(StringOf(permVect.probB^[j] : 5 : 3));
			DisplayLn;
		end; { for j }
		DisplayLn;
		DisplayLn;

		DisplayStringLn('   R^2          =      >   Prob ');
		DisplayStringLn('------------------------  ------');
		DisplayString(StringOf(rCarreOriginal : 10 : 8));
		DisplayString('  ');
		DisplayString(StringOf(permVect.egr : 5));
		DisplayString('  ');
		DisplayString(StringOf(permVect.pgr : 5));
		DisplayString('  ');
		DisplayString(StringOf(permVect.probR : 6 : 4));
		DisplayLn;
		DisplayLn;
	end; { AfficherResultatsComplets }

	procedure FaireTestSimple (nbPerm: TInteger;
									y, x: TMatrixPtr;
									progress: ProgressInfoPtr);
		var
			bOriginal, tOriginal: TMatrixPtr;
			rCarreOriginal: TReal;
			permVect: TPermuteVect;
			oldWindow: WindowPtr;
	begin
		bOriginal := nil;
		tOriginal := nil;
		rCarreOriginal := 0;

		CreatePermuteVect(x^.p, permVect);

		GetPort(oldWindow);

		SetDWindow(progress^.displayWindow);

		DisplayStringLn('-----------');
		DisplayStringLn('Simple test');
		DisplayStringLn('-----------');
		DisplayLn;

		AfficherEnTete(nbPerm, y^.permMeth, y^.root, GetColName(y, 1), 0, 0);

		RegressionMultiplePermutee(nbPerm, y, x, bOriginal, tOriginal, rCarreOriginal, permVect, progress);

		SetDWindow(progress^.displayWindow);
		AfficherResultatsComplets(nbPerm, permVect, x, bOriginal, rCarreOriginal);

		SetPort(oldWindow);

		DisposeMatrix(bOriginal);
		DisposeMatrix(tOriginal);
		DisposePermuteVect(permVect);
	end; { FaireTestSimple }

	procedure FaireBackward (nbPerm: TInteger;
									y, x: TMatrixPtr;
									pToRemove: TReal;
									isStepwise: Boolean;
									variablesEliminees: TBoolVectorPtr;
									progress: ProgressInfoPtr);

		var
			b, t, xBackward, xTemp, matTemp: TMatrixPtr;
			permVect: TPermuteVect;
			numIncluded, etape: TInteger;
			rCarre: TReal;
			i, plusGrandIndex: TInteger;
			plusGrandeProb, plusPetiteProb, probMinPossible, probBonf: TReal;
			toutesVariablesSignificatives: Boolean;
			oldWindow: WindowPtr;

	begin
		b := nil;
		t := nil;
		xBackward := nil;
		xTemp := nil;
		toutesVariablesSignificatives := False;
		numIncluded := x^.p; { number of variables included in procedure at the start }
		probMinPossible := 1 / (nbPerm + 1); { Probabilit minimum possible }

		GetPort(oldWindow);
		SetDWindow(progress^.displayWindow);


		if not isStepwise then begin
			DisplayStringLn('------------------------------');
			DisplayStringLn('Backward elimination procedure');
			DisplayStringLn('------------------------------');
			DisplayLn;

			AfficherEnTete(nbPerm, y^.permMeth, y^.root, GetColName(y, 1), 0, pToRemove);
		end;{ if }


{ Copier "x" dans "xBackward", car on effacera potentiellement des colonnes de xBackward }
		DuplicateMatrix(x, xBackward, 0, 0, true); { Copier avec les noms }

		etape := 0;

		repeat
			CreatePermuteVect(numIncluded, permVect);

			etape := etape + 1;
			DisplayLn;
			DisplayStringLn(StringOf('Step ', etape : 1));
			DisplayLn;

			RegressionMultiplePermutee(nbPerm, y, xBackward, b, t, rCarre, permVect, progress);

			SetDWindow(progress^.displayWindow);

{ Trouver quelle variable de xBackward a la probabilit la plus leve }
			plusGrandeProb := -1;
			plusPetiteProb := 2;
			plusGrandIndex := 1;
			for i := 1 to numIncluded do begin
				if permVect.probB^[i] > plusGrandeProb then begin
					plusGrandIndex := i;
					plusGrandeProb := permVect.probB^[i];
				end; { if }

				if permVect.probB^[i] < plusPetiteProb then
					plusPetiteProb := permVect.probB^[i];

			end; { for i }

			probBonf := pToRemove / numIncluded; { Correction de Bonferroni }
			toutesVariablesSignificatives := plusGrandeProb <= probBonf;
			if not toutesVariablesSignificatives then
				if (probMinPossible > probBonf) then begin { Oups! On ne peux pas aller plus bas dans les probabilits! }
					DisplayLn;
					DisplayStringLn('*** Caution! ');
					DisplayStringLn(StringOf('    Minimal possible probability is ', probMinPossible : 5 : 3));
					DisplayStringLn(StringOf('    Bonferonni-corrected p-to-remove is ', probBonf : 5 : 3));
					DisplayStringLn('*** Procedure will now continue but you should try again with more permutations.');
					DisplayLn;
				end; { if }

			AfficherResultatsComplets(nbPerm, permVect, xBackward, b, rCarre);

			DisposeMatrix(b);
			DisposeMatrix(t);
			DisposePermuteVect(permVect);

			numIncluded := numIncluded - 1;

			if (numIncluded > 0) and not toutesVariablesSignificatives then begin
{ on continue en enlevant la variable la moins significative, }
{ c'est la variable "plusGrandIndex" }
{ Copier xBackward dans xTemp }
				DuplicateMatrix(xBackward, xTemp, 0, plusGrandIndex, true);

{ Si on fait une stepwise, il faut aussi indiquer que cette variable }
{ "plusGrandIndex" a t limine: }
				if isStepwise then
					variablesEliminees^[plusGrandIndex] := true;

{ changer les valeurs de xBackward et xTemp, car ce qui nous intresse se trouve dans xTemp }
{ mais on le veut dans xBackward }
				matTemp := xTemp;
				xTemp := xBackward;
				xBackward := matTemp;

{ Disposer de xTemp pour rcuprer la mmoire }
				DisposeMatrix(xTemp);
{ matTemp n'est qu'un pointeur, pas de mmoire alloue pour cette variable, rien  faire avec }
			end;  { if }

		until (numIncluded = 0) or toutesVariablesSignificatives;

		DisposeMatrix(xBackward);

		if not isStepwise then begin
			DisplayStringLn('----------------------------------');
			DisplayStringLn('End Backward elimination procedure');
			DisplayStringLn('----------------------------------');
			DisplayLn;
		end;{ if }

		SetPort(oldWindow);

	end;{ procedure FaireBackward }

	procedure FaireForward (nbPerm: TInteger;
									y, x: TMatrixPtr;
									pToEnter, pToRemove: TReal; { pToRemove dans le cas de stepwise }
									isStepwise: Boolean;
									progress: ProgressInfoPtr);

		var
			b, t, xForward: TMatrixPtr;
			permVect: TPermuteVect;
			j, etape, quelleVariable, jj, nbVarEliminees, ii, {}
			varDansX, prochVariable, jjDansX, prochVariableDansX: TInteger;
			rCarre, probMinPossible, probBonf, rCarreEtapePrecedente: TReal;
			variablesSelectionnees, elimineParBackward: TBoolVectorPtr;
			probRCarre, probB, deltaRCarre: TRealVectorPtr;
			variableXDansXForward: TIntegerVectorPtr;
			premiereVariable: Boolean;
			oldWindow: WindowPtr;

		function TrouverVariableAEntrer (n: TInteger;
										varSelectionnees: TBoolVectorPtr;
										probRCarre, probB, deltaRCarre: TRealVectorPtr;
										probBonf: TReal): Tinteger;
			var
				i, result: TInteger;
				pRCarreMin, pBMin, deltaRCarreMax: TReal;
				autresVariables: Boolean;
		begin
			result := 0;
{ Toutes les variables ont-elles t slectionnes? }
			for i := 1 to n do
				if not (varSelectionnees^[i]) then
					result := result + 1;

			if result > 0 then begin { on a au moins une variable  vrifier }
				pRCarreMin := 1;
				pBMin := 1; { On va aussi vrifier si le b de cette variable au R Carr significatif est aussi significatif }
				for i := 1 to n do
					if not varSelectionnees^[i] then
						if probRCarre^[i] < pRCarreMin then begin
							pRCarreMin := probRCarre^[i];
							pBMin := probB^[i];
							result := i;
						end; { if }

{ On a trouv la plus petite probabilit de r carr, et celle du b de la mme variable. }
{ Celle du R Carr est-elle significative selon Bonferonni? }
{ (on verra pour celle du b plus bas) }
				if pRCarreMin > probBonf then
					result := 0;

				if result > 0 then begin
{    D 'autres variables ont-elles cette probabilit de r carr? }
					autresVariables := false;
					for i := 1 to n do
						if i <> result then { on exclue la variable choisie bien sr! }
							if not varSelectionnees^[i] then
								if probRCarre^[i] = pRCarreMin then
									autresVariables := true; { Au moins une autre variable a cette valeur de p(r^2) }

					if autresVariables then begin { 2e critre, probabilit du b }
						pBMin := 1;

						for i := 1 to n do
							if not varSelectionnees^[i] then
								if probRCarre^[i] = pRCarreMin then
									if probB^[i] < pBMin then begin
										pBMin := probB^[i];
										result := i;
									end; { if }

{ On a trouv la plus petite probabilit de b. Cette probabilit est-elle significative selon Bonferonni? }
						if pBMin > probBonf then
							result := 0; { un r^2 peut tre significatif, mais la variable qui a t ajoute peut ne pas l'tre }

						if result > 0 then begin
{    D 'autres variables ont-elles cette probabilit de b? }
							autresVariables := false;
							for i := 1 to n do
								if i <> result then { on exclue la variable choisie bien sr! }
									if not varSelectionnees^[i] then
										if probRCarre^[i] = pRCarreMin then
											if probB^[i] = pBMin then
												autresVariables := true; { Au moins une autre variable a cette valeur de p(b) }

							if autresVariables then begin { 3e critre, delta-r^2 }
								deltaRCarreMax := 0;

								for i := 1 to n do
									if not varSelectionnees^[i] then
										if probRCarre^[i] = pRCarreMin then
											if probB^[i] = pBMin then
												if deltaRCarre^[i] > deltaRCarreMax then begin
													deltaRCarreMax := deltaRCarre^[i];
													result := i;
												end; { if }
							end; { if autresVariables }

						end; { if result }

					end { if autresVariables }
					else { Cas spcial s'il n'y a pas d'autre variable }
				{ Dans ce cas, il faut aussi vrifier la probabilit du b }
						if pBMin > ProbBonf then
							result := 0;

				end; { if result }

			end; { if result }

			TrouverVariableAEntrer := result;
		end; { TrouverVariableAEntrer }

		procedure CopierColonne (xCol, xForwardCol: TInteger);
			var
				ii: TInteger;
		begin
		{ Copier dans xForward }
			for ii := 1 to x^.n do
				SetElement(xForward, ii, xForwardCol, GetElement(x, ii, xCol));
			SetColName(xForward, xForwardCol, GetColName(x, xCol));
		{ Copier l'information de permutations }
		end; { CopierColonne }

		procedure AfficherEnTeteForward;
		begin
			DisplayStringLn('Testing the following variables for inclusion in model:');
			DisplayStringLn('Variable    Std. b        p(b)     r^2       p(r^2)');
			DisplayStringLn('----------  ------------------    ------    --------');
		end; { AfficherEnTeteForward }

		procedure AfficherResultatUneVariable (nbPerm: TInteger;
										permVect: TPermuteVect;
										x, b: TMatrixPtr;
										rCarre: TReal);
			var
				i, nbDecimales: TInteger;
				ts: TString;
		begin
{$ifc undefined THINK_PASCAL}
			nbDecimales := system.Trunc(Ln(nbPerm) / Ln(10)) + 1;
{$elsec}
			nbDecimales := Trunc(Ln(nbPerm) / Ln(10)) + 1;
{$endc}
			ts := GetColName(x, x^.p);
			DisplayString(ts);
			for i := Length(ts) to 10 do { pad with spaces }
				DisplayString(' ');
			if GetElement(b, x^.p, 1) > 0 then
				DisplayString(' ');
			DisplayString(StringOf(GetElement(b, x^.p, 1) : 6 : 4));
			DisplayString('  ');
			for i := nbDecimales to 7 do
				DisplayString(' ');
			DisplayString(StringOf(permVect.probB^[x^.p] : nbDecimales + 2 : nbDecimales));
			DisplayString('    ');
			DisplayString(StringOf(rCarre : 6 : 4));
			DisplayString('  ');
			for i := nbDecimales to 7 do
				DisplayString(' ');
			DisplayString(StringOf(permVect.probR : nbDecimales + 2 : nbDecimales));
			DisplayLn;
		end; { AfficherResultatUneVariable }

	begin
		b := nil;
		t := nil;
		xForward := nil;
{ Vecteurs utiliss pour les dcisions en Forward selection, voir ci-dessous }
		probRCarre := TRealVectorPtr(NewPtrClear(x^.p * SizeOf(TReal)));
		probB := TRealVectorPtr(NewPtrClear(x^.p * SizeOf(TReal)));
		deltaRCarre := TRealVectorPtr(NewPtrClear(x^.p * SizeOf(TReal)));

		probMinPossible := 1 / (nbPerm + 1); { Probabilit minimum possible }

		GetPort(oldWindow);
		SetDWindow(progress^.displayWindow);


{ Vecteur de variables dj slectionnes }
		variablesSelectionnees := TBoolVectorPtr(NewPtrClear(x^.p * SizeOf(Boolean)));
		Assert(variablesSelectionnees <> nil);
		if variablesSelectionnees <> nil then
			for j := 1 to x^.p do
				variablesSelectionnees^[j] := false; { Aucune variable slectionne au dpart }

		if isStepwise then begin
{ Vecteur de variables limines par la backward lors d'une stepwise }
			elimineParBackward := TBoolVectorPtr(NewPtrClear(x^.p * SizeOf(Boolean)));
			Assert(elimineParBackward <> nil);
			for j := 1 to x^.p do
				elimineParBackward^[j] := false; { Aucune variable limine au dpart }

{ Vecteur d'entiers utilis pour se souvenir de la concordance entre les variables }
{ dans X et XForward }
			variableXDansXForward := TIntegerVectorPtr(NewPtrClear(x^.p * SizeOf(TInteger)));
			Assert(variableXDansXForward <> nil);
			for j := 1 to x^.p do
				variableXDansXForward^[j] := 0;
		end; { if isStepwise }

		if not isStepwise then begin
			DisplayStringLn('---------------------------');
			DisplayStringLn('Forward Selection procedure');
			DisplayStringLn('---------------------------');
			DisplayLn;

			AfficherEnTete(nbPerm, y^.permMeth, y^.root, GetColName(y, 1), pToEnter, 0);
		end;{ if }


{ Copier "x" dans "xForward" }
		DuplicateMatrix(x, xForward, 0, 0, true); { Copier avec les noms }
{ Note: en fait, on se fout un peu de ce qu'il y a dans xForward  ce moment, car }
{ cette matrice sera remplie au fur et  mesure plus bas; on modifiera la valeur }
{ "xForward^.p" pour faire fonctionner la procdure de rgression multiple correctement }

		etape := 0;

		rCarreEtapePrecedente := 0;

		repeat
			etape := etape + 1;
			CreatePermuteVect(etape, permVect);

			DisplayLn;
			DisplayString(StringOf('Step ', etape : 1, '. '));
{ Afficher les variables dja slectionnes }
			if etape = 1 then
				DisplayStringLn('No variables selected.')
			else begin
				DisplayString('Variables already included in model are: ');
				for j := 1 to xForward^.p do begin
					if j > 1 then
						DisplayString(', ');
					DisplayString(GetColName(xForward, j));
				end; { for j }
				DisplayLn;
			end; { else }
			DisplayLn;
			AfficherEnTeteForward;

{ Pour effectuer une forward selection, il faut: }
{ - Copier tour  tour les variables restantes dans la "etape"-ime colonne de xForward }
{ - Passer la matrice xForward  la rgression multiple }
{ - Noter quelle variable cause le r^2 le plus significatif: }
{   * S'il y en a plusieurs, choisir celle qui a le coefficient b le plus significatif }
{   * S'il y en a encore plusieurs, choisir celle qui donne l'augmentation du r^2 (delta r^2) }
{     la plus grande. }

			xForward^.p := etape; { Rgler le nombre de colonnes de xForward }

			for j := 1 to x^.p do begin { Boucle sur toutes les variables }
				if not variablesSelectionnees^[j] then begin { Essayer avec cette variable }

{ Copier la j-me variable de x dans la etape-ime colonne de xForward }
					CopierColonne(j, etape);

{ Effectuer la rgression multiple avec notre matrice xForward modifie }
					ClearPermuteVect(etape, permVect);
					RegressionMultiplePermutee(nbPerm, y, xForward, b, t, rCarre, permVect, progress);
					SetDWindow(progress^.displayWindow);

{ Conserver les rsultats de ce calcul dans les trois vecteurs de comparaison }
					probRCarre^[j] := permVect.probR;
					probB^[j] := permVect.probB^[etape]; { Seule la etape-ime colonne nous intresse }
					deltaRCarre^[j] := rCarre - rCarreEtapePrecedente;

					AfficherResultatUneVariable(nbPerm, permVect, xForward, b, deltaRCarre^[j]);

				end; { if not variablesSelectionnees }
			end; { for j }

{ Trouver quelle variable restante doit tre slectionne }
			probBonf := pToEnter / etape; { Correction de Bonferroni }
			quelleVariable := TrouverVariableAEntrer(x^.p, variablesSelectionnees, probRCarre, probB, deltaRCarre, probBonf);

			if quelleVariable > 0 then begin
{ Indiquons que la variable "quelleVariable" dans x correspond  la "etape"-ienne dans xForward }
				if isStepwise then
					variableXDansXForward^[quelleVariable] := etape;

{ On a trouv une variable  inclure dans le modle. Ajoutons-l  xForward: }
				CopierColonne(quelleVariable, etape);
				variablesSelectionnees^[quelleVariable] := true;

				rCarreEtapePrecedente := rCarreEtapePrecedente + deltaRCarre^[quelleVariable];
				DisplayLn;
				DisplayStringLn(StringOf('*** Variable "', GetColName(x, quelleVariable), '" was found significant and is added to the model.'));
				DisplayLn;

				if isStepwise and (etape > 1) then begin
{ Stepwise regression: }
{  chaque tape, on value la matrice xForward par un backward elimination. }
{ Si une ou des variables sont limines par cette backward, on les retire de xForward. }

{ Restaurer les variables limines par le backward prcdent, s'il y a lieu }
					if isStepwise then
						for jj := 1 to x^.p do begin
							if elimineParBackward^[jj] then begin
{ la jj-ime variable de xForward a t limine par la backward prcdente. }
{ Il faut donc trouver  quelle variable de x cette variable de xForward correspond... }
								varDansX := 0;
								for ii := 1 to x^.p do
									if variableXDansXForward^[ii] = jj then
										varDansX := ii;
{ et la rendre  nouveau disponible }
								Assert(varDansX > 0);
								variablesSelectionnees^[varDansX] := False;

							end; { if }
						end; { for jj }

					nbVarEliminees := 0;
					for jj := 1 to xForward^.p do
						elimineParBackward^[jj] := False; { Aucune variable limine pour le moment }

					DisplayStringLn('+++ Assessing significance of the model using a backward elimination procedure');
					FaireBackward(nbPerm, y, xForward, pToRemove, IS_STEPWISE, elimineParBackward, progress);

					for jj := 1 to xForward^.p do
						if elimineParBackward^[jj] then
							nbVarEliminees := nbVarEliminees + 1;

					if nbVarEliminees > 0 then begin
{ Au moins une variable a t exclue par la procdure Backward. }
{ En informer l'usager }
						DisplayLn;
						DisplayString('Variable');
						if nbVarEliminees > 1 then
							DisplayString('s');
						DisplayString(' ');
						premiereVariable := true;
						for jj := 1 to xForward^.p do
							if elimineParBackward^[jj] then begin
								if premiereVariable then
									premiereVariable := False
								else
									DisplayString(', ');
								DisplayString(GetColName(xForward, jj));
							end; { if }
						if nbVarEliminees = 1 then
							DisplayString(' was ')
						else
							DisplayString(' were ');
						DisplayStringLn('found not significant by the backward elimination procedure.');
						if nbVarEliminees = 1 then
							DisplayString('This variable is ')
						else
							DisplayString('These variables are ');
						DisplayStringLn('thus excluded from the next step in forward selection.');
						DisplayLn;

{ Puis il faut la/les enlever de xForward, et ajuster etape et xForward^.p en consquence. }
						jj := 1;
						while jj <= xForward^.p do begin
{ Si la jj-ime colonne a t limine, copier la prochaine colonne non-limine }
							if elimineParBackward^[jj] then begin
								prochVariable := jj + 1;
								while (prochVariable <= xForward^.p) and elimineParBackward^[prochVariable] do
									prochVariable := prochVariable + 1;
{ La variable  copier dans la jj-ime colonne se trouve dans prochVariable }
								if (prochVariable <= xForward^.p) then begin
									for ii := 1 to xForward^.n do
										SetElement(xForward, ii, jj, GetElement(xForward, ii, prochVariable));
									SetColName(xForward, jj, GetColName(xForward, prochVariable)); { Copier le nom de la variable }
{elimineParBackward^[prochVariable] := True;}
 { on marque 'true' pour ne plus copier }

{ Il reste maintenant  modifier les positions dans xForward: }
{ En bref, on dsire que l'index dans x de la variable jj de xForward soit mis  zro }
{ et que l'index dans x de la variable prochVariable soit mis  celui de jj }
{ 1) Trouver ces deux index: }
									jjDansX := 0;
									prochVariableDansX := 0;
									for ii := 1 to x^.p do begin
										if variableXDansXForward^[ii] = jj then
											jjDansX := ii;
										if variableXDansXForward^[ii] = prochVariable then
											prochVariableDansX := ii;
									end; { for ii }
{ 2) Faire la manipulation dcrite plus haut }
									Assert(jjDansX > 0);
									Assert(prochVariableDansX > 0);
									variableXDansXForward^[jjDansX] := 0;
									variableXDansXForward^[prochVariableDansX] := jj;
								end; { if prochVariable }
{ Voil! Ouf! Fini le "bookkeeping"! }
							end; { if elimineParBackward }
							jj := jj + 1;
						end; { while jj }

{ On a maintenant une matrice xForward qui contient seulement les variables non- }
{ limines par la backward. Modifions le vecteur "varSelectionnees" pour forcer }
{ le programme  ne pas utiliser les variables limines }
						for jj := 1 to xForward^.p do
							if elimineParBackward^[jj] then begin
{ La jj-ime variable de xForward a t limine. }
{ Mais  quelle variable de x correspond-t-elle? }
								varDansX := 0;
								for ii := 1 to x^.p do
									if variableXDansXForward^[ii] = jj then
										varDansX := ii;
{ Rendre cette variable dans x non-disponible en prtendant qu'elle }
{ est dj slectionne, ce qui aura pour effet de l'ignorer lors de la }
{ prochaine forward }
								Assert(varDansX > 0);
								variablesSelectionnees^[varDansX] := True;
							end; { if }

						Assert(nbVarEliminees < etape);
						etape := etape - nbVarEliminees;
						xForward^.p := xForward^.p - nbVarEliminees;

					end { nbVarEliminees > 0 }

					else begin { aucune variable limine }
						DisplayLn;
						DisplayStringLn('No variables were removed by the backward elimination procedure.');
						DisplayLn;
					end; { else }

				end; { if isStepwise }

			end; { if quelleVariable }

			DisposeMatrix(b);
			DisposeMatrix(t);
			DisposePermuteVect(permVect);

		until (quelleVariable = 0);

		DisplayLn;
		DisplayStringLn('No more variables can be added to the model.');
		DisplayLn;

		xForward^.p := x^.p; { Rgler le nombre de colonnes de xForward  sa valeur originale }
		DisposeMatrix(xForward);
		DisposePtr(Ptr(variablesSelectionnees));
		DisposePtr(Ptr(probRCarre));
		DisposePtr(Ptr(probB));
		DisposePtr(Ptr(deltaRCarre));

		if isStepwise then begin
			DisposePtr(Ptr(elimineParBackward));
			DisposePtr(Ptr(variableXDansXForward));
		end; { if isStepwise }

		if not isStepwise then begin
			DisplayStringLn('-------------------------------');
			DisplayStringLn('End Forward Selection procedure');
			DisplayStringLn('-------------------------------');
			DisplayLn;
		end; { if }

		SetPort(oldWindow);

	end;{ procedure FaireForward }

	function FaireTests (ti: TestInfoPtr): LongInt;
		var
			x, y: TMatrixPtr;
			j: TInteger;
	begin

		Assert(ti^.mat <> nil);
		Assert(ti^.whichXVariables <> nil);
		Assert(ti^.whichYVariable <> nil);

		x := nil;
		y := nil;

		SubMatrix(ti^.mat, x, ti^.whichXVariables, true);

		SubMatrix(ti^.mat, y, ti^.whichYVariable, true);

{$ifc do_debugfiles}
		writeln(debug2, 'Vecteur y');
		for j := 1 to y^.n do
			writeLn(debug2, GetElement(y, j, 1) : 8 : 6);
		writeln(debug2);
		writeln(debug2, 'Matrice X');
		for i := 1 to x^.p do begin
			for j := 1 to x^.n do
				write(debug2, GetElement(x, j, i) : 8 : 6, '    ');
			writeln(debug2);
		end;
{$endc}

{ Centrage et rduction de x et y }
		for j := 1 to x^.p do
			CentrerEtReduire(x, j);
		CentrerEtReduire(y, 1);

{ Mise en place des mthodes et outils de permutation }
		y^.permMeth := ti^.permMeth;
		y^.root := ti^.root;
		CreerVecteurIndirection(y);

{ Should do a memory check for x, y... }

		InitRan3(-1, 100); { Initialiser le gnrateur de nombres alatoires }

		case ti^.testMeth of
			simple: 
				FaireTestSimple(ti^.nbPerm, y, x, ti^.progress);

			backward: 
				FaireBackward(ti^.nbPerm, y, x, ti^.pToRemove, not (IS_STEPWISE), nil, ti^.progress);

			forward: 
				FaireForward(ti^.nbPerm, y, x, ti^.pToEnter, 0, not (IS_STEPWISE), ti^.progress);

			stepwise:  begin
				DisplayStringLn('-----------------------------');
				DisplayStringLn('Stepwise regression procedure');
				DisplayStringLn('-----------------------------');
				DisplayLn;

				AfficherEnTete(ti^.nbPerm, y^.permMeth, y^.root, GetColName(y, 1), ti^.pToEnter, ti^.pToRemove);

				FaireForward(ti^.nbPerm, y, x, ti^.pToEnter, ti^.pToRemove, IS_STEPWISE, ti^.progress);

				DisplayStringLn('---------------------------------');
				DisplayStringLn('End Stepwise regression procedure');
				DisplayStringLn('---------------------------------');
				DisplayLn;

			end;
			otherwise
				;
		end; { case testMeth }

{ Once this procedure is completed, we need to dispose of especially of the two boolean }
{ vectors that came with it }
		DisposePtr(Ptr(ti^.whichXVariables));
		DisposePtr(Ptr(ti^.whichYVariable));

		DisposeMatrix(x);
		DisposeMatrix(y);

		FaireTests := 1; { return value > 0 }

	end; { FaireTest }

end. { Unit DoTests }