unit InterfacesUI;

{Unit for making Think Pascal work with Universal Interfaces}
{All units that are not in Interfaces.lib are not reflected here.}
{This implements changes to the core Toolbox, excluding most System7-}
{specific units. You have to inlude them in your project anyway, so}
{you can just as well make the changes in the interface files as}
{necessary.}

{ Modified 16/03/98 by PhC: added "qd" object to make source-code compatibility }
{ with CW Pascal for all QuickDraw variables in Universal Interfaces, e.g. }
{ "thePort" is now "qd.thePort". Just use "new(qd)" at the beginning of your program. }
{ 29/06/98: removed that modification (set compiler switch Use_QDGlobalObjects to false }
{ as I can't make it work for now }

{ Modified 08/05/98 PhC: added a couple of UPP procedures for modal dialog filters }

{ Modified 29/07/98 PhC: added a check for SANE functions. Set the compiler variable}
{ USE_SANE to true in the Project->Compile Options... box to use the conversions. }

{ Modified 31/07/98 PhC: changed *Ref declarations to *Handle (e.g. ControlRef -> }
{ ControlHandle) since they are now obsolete and have reverted back to being *Handles }
{ now that Copland is scrapped and STRICT_WINDOWS does not exist anymore... }

{ Modified 13/08/98 PhC: added Finder interface calls and constants }

interface

{$SETC GENERATINGPOWERPC := FALSE}
{$IFC UNDEFINED POWERPC}
{$SETC POWERPC := FALSE}
{$ENDC}

{$IFC UNDEFINED PROCTYPE}
{$SETC PROCTYPE := 0}
{$ENDC}

{$IFC UNDEFINED GENERATINGCFM}
{$SETC GENERATINGCFM := 0}
{$ENDC}

{$IFC UNDEFINED Use_QDGlobalObjects}
{$SETC Use_QDGlobalObjects := FALSE}
{$ENDC}

{Controls}
	const
		kControlNoPart = 0;
		kControlLabelPart = 1;
		kControlMenuPart = 2;
		kControlTrianglePart = 4;
		kControlButtonPart = 10;
		kControlCheckBoxPart = 11;
		kControlRadioButtonPart = 11;
		kControlUpButtonPart = 20;
		kControlDownButtonPart = 21;
		kControlPageUpPart = 22;
		kControlPageDownPart = 23;
		kControlIndicatorPart = 129;
		kControlDisabledPart = 254;
		kControlInactivePart = 255;
	type
		ControlRef = ControlHandle;
		ConstStr255Param = Str255;
		WindowRef = WindowPtr;
		ControlActionUPP = ProcPtr;
		NewControlActionProc = ProcPtr;
		UniversalProcPtr = ProcPtr;
		ModalFilterUPP = ProcPtr;
		ModalFilterProcPtr = ProcPtr;  { FUNCTION ModalFilter(theDialog: DialogPtr; VAR theEvent: EventRecord; VAR itemHit: DialogItemIndex): BOOLEAN; }
		ListSearchUPP = ProcPtr;
		ControlDefProcPtr = ProcPtr;  { FUNCTION ControlDef(varCode: SInt16; theControl: ControlHandle; message: ControlDefProcMessage; param: SInt32): SInt32; }
		ControlDefUPP = UniversalProcPtr;

{$ifc Use_QDGlobalObjects}
		QDGlobals = object
				function randSeed: LONGINT;
				function screenBits: BitMap;
				function arrow: Cursor;
				function dkGray: Pattern;
				function ltGray: Pattern;
				function gray: Pattern;
				function black: Pattern;
				function white: Pattern;
				function thePort: GrafPtr;
			end; { QDGlobals }
	var
		qd: QDGlobals;
{$endc}

	procedure SetControlTitle (theControl: ControlHandle;
									title: ConstStr255Param);
	inline
		$A95F;
	procedure GetControlTitle (theControl: ControlHandle;
									var title: Str255);
	inline
		$A95E;
	procedure UpdateControls (theWindow: WindowPtr;
									updateRgn: RgnHandle);
	inline
		$A953;
	procedure SetControlValue (theControl: ControlHandle;
									theValue: INTEGER);
	inline
		$A963;
	function GetControlValue (theControl: ControlHandle): INTEGER;
	inline
		$A960;
	procedure SetControlMinimum (theControl: ControlHandle;
									minValue: INTEGER);
	inline
		$A964;
	function GetControlMinimum (theControl: ControlHandle): INTEGER;
	inline
		$A961;
	procedure SetControlMaximum (theControl: ControlHandle;
									maxValue: INTEGER);
	inline
		$A965;
	function GetControlMaximum (theControl: ControlHandle): INTEGER;
	inline
		$A962;
	function GetAuxiliaryControlRecord (theControl: ControlHandle;
									var acHndl: AuxCtlHandle): BOOLEAN;
	inline
		$AA44;
	procedure SetControlReference (theControl: ControlHandle;
									data: LONGINT);
	inline
		$A95B;
	function GetControlReference (theControl: ControlHandle): LONGINT;
	inline
		$A95A;
	procedure SetControlAction (theControl: ControlHandle;
									actionProc: ControlActionUPP);
	inline
		$A96B;
	function GetControlAction (theControl: ControlHandle): ControlActionUPP;
	inline
		$A96A;
	procedure SetControlColor (theControl: ControlHandle;
									newColorTable: CCTabHandle);
	inline
		$AA43;
	function GetControlVariant (theControl: ControlHandle): INTEGER;
	inline
		$A809;
	function NewControlDefProc (userRoutine: ControlDefProcPtr): ControlDefUPP;
	inline
		$2E9F;
{Dialogs}
	type
		DialogRef = DialogPtr;
		NewUserItemProc = ProcPtr;
		UserItemProcPtr = ProcPtr;
	procedure DisposeDialog (theDialog: DialogPtr);
	inline
		$A983;
	procedure UpdateDialog (theDialog: DialogPtr;
									updateRgn: RgnHandle);
	inline
		$A978;
	procedure GetDialogItem (theDialog: DialogPtr;
									itemNo: INTEGER;
									var itemType: INTEGER;
									var item: Handle;
									var box: Rect);
	inline
		$A98D;
	procedure SetDialogItem (theDialog: DialogPtr;
									itemNo: INTEGER;
									itemType: INTEGER;
									item: Handle; {CONST}
									var box: Rect);
	inline
		$A98E;
	procedure HideDialogItem (theDialog: DialogPtr;
									itemNo: INTEGER);
	inline
		$A827;
	procedure ShowDialogItem (theDialog: DialogPtr;
									itemNo: INTEGER);
	inline
		$A828;
	procedure SelectDialogItemText (theDialog: DialogPtr;
									itemNo: INTEGER;
									strtSel: INTEGER;
									endSel: INTEGER);
	inline
		$A97E;
	procedure GetDialogItemText (item: Handle;
									var text: Str255);
	inline
		$A990;
	procedure SetDialogItemText (item: Handle;
									text: ConstStr255Param);
	inline
		$A98F;
	function FindDialogItem (theDialog: DialogPtr;
									thePt: Point): INTEGER;
	inline
		$A984;

	function SetDialogDefaultItem (theDialog: DialogPtr;
									newItem: INTEGER): OSErr;
	inline
		$303C, $0304, $AA68;

	function SetDialogCancelItem (theDialog: DialogPtr;
									newItem: INTEGER): OSErr;
	inline
		$303C, $0305, $AA68;

	function SetDialogTracksCursor (theDialog: DialogPtr;
									tracks: BOOLEAN): OSErr;
	inline
		$303C, $0306, $AA68;

	function NewColorDialog (dStorage: univ Ptr; {CONST}
									var boundsRect: Rect;
									title: ConstStr255Param;
									visible: BOOLEAN;
									procID: INTEGER;
									behind: WindowPtr;
									goAwayFlag: BOOLEAN;
									refCon: LONGINT;
									items: Handle): DialogPtr;
	inline
		$AA4B;
	function GetAlertStage: INTEGER;
	inline
		$3EB8, $0A9A;			{ MOVE.W $0A9A,(SP) }

	procedure ResetAlertStage;
	inline
		$4278, $0A9A;
	procedure DialogCut (theDialog: DialogPtr);
	procedure DialogPaste (theDialog: DialogPtr);
	procedure DialogCopy (theDialog: DialogPtr);
	procedure DialogDelete (theDialog: DialogPtr);
	procedure SetDialogFont (value: INTEGER);
	inline
		$31DF, $0AFA;			{ MOVE.W (SP)+,$0AFA }

{Events}
	function KeyTranslate (transData: univ Ptr;
									keycode: INTEGER;
									var state: LONGINT): LONGINT;
	inline
		$A9C3;

{Finder interface; System 7 and later}
	const
		kAlignNone = $00;
		kAlignVerticalCenter = $01;
		kAlignTop = $02;
		kAlignBottom = $03;
		kAlignHorizontalCenter = $04;
		kAlignAbsoluteCenter = $05;
		kAlignCenterTop = $06;
		kAlignCenterBottom = $07;
		kAlignLeft = $08;
		kAlignCenterLeft = $09;
		kAlignTopLeft = $0A;
		kAlignBottomLeft = $0B;
		kAlignRight = $0C;
		kAlignCenterRight = $0D;
		kAlignTopRight = $0E;
		kAlignBottomRight = $0F;

		kTransformNone = $00;
		kTransformDisabled = $01;
		kTransformOffline = $02;
		kTransformOpen = $03;
		kTransformLabel1 = $0100;
		kTransformLabel2 = $0200;
		kTransformLabel3 = $0300;
		kTransformLabel4 = $0400;
		kTransformLabel5 = $0500;
		kTransformLabel6 = $0600;
		kTransformLabel7 = $0700;
		kTransformSelected = $4000;
		kTransformSelectedDisabled = $4001;
		kTransformSelectedOffline = $4002;
		kTransformSelectedOpen = $4003;

	function PlotIconID (var theRect: Rect;
									alignment, transform, theResID: Integer): OSErr;
	inline
		$303C, $0500, $ABC9;

{Lists}
	procedure LSetDrawingMode (drawIt: BOOLEAN;
									lHandle: ListHandle);
	inline
		$3F3C, $002C, $A9E7;
	procedure LGetCellDataLocation (var offset: INTEGER;
									var len: INTEGER;
									theCell: Cell;
									lHandle: ListHandle);
	inline
		$3F3C, $0034, $A9E7;
	function NewListDefProc (userRoutine: ProcPtr): UniversalProcPtr;
	inline
		$2E9F;
	function NewListSearchProc (userRoutine: ProcPtr): ListSearchUPP;
	inline
		$2E9F;

{Memory}
	function ApplicationZone: THz;
	inline
		$2EB8, $02AA;			{ MOVE.L $02AA,(SP) }
	function TempNewHandle (logicalSize: Size;
									var resultCode: OSErr): Handle;
	inline
		$3F3C, $001D, $A88F;
	function TempMaxMem (var grow: Size): Size;
	inline
		$3F3C, $0015, $A88F;
	function TempFreeMem: LONGINT;
	inline
		$3F3C, $0018, $A88F;
 {  Temporary Memory routines renamed, but obsolete, in System 7.0 and later.  }
	procedure TempHLock (h: Handle;
									var resultCode: OSErr);
	inline
		$3F3C, $001E, $A88F;
	procedure TempHUnlock (h: Handle;
									var resultCode: OSErr);
	inline
		$3F3C, $001F, $A88F;
	procedure TempDisposeHandle (h: Handle;
									var resultCode: OSErr);
	inline
		$3F3C, $0020, $A88F;
	function TempTopMem: Ptr;
	inline
		$3F3C, $0016, $A88F;
	procedure ReserveMem (cbNeeded: Size);
	inline
		$201F, $A040;
	procedure ReserveMemSys (cbNeeded: Size);
	inline
		$201F, $A440;
	procedure DisposePtr (p: Ptr);		{Unnecessary?}
	inline
		$205F, $A01F;
	procedure DisposeHandle (h: Handle);	{Unnecessary?}
	inline
		$205F, $A023;
	procedure ReallocateHandle (h: Handle;
									byteCount: Size);
	inline
		$201F, $205F, $A027;
 {Menus}
	type
		MenuRef = MenuHandle;
	procedure AppendResMenu (theMenu: MenuHandle;
									theType: ResType);
	inline
		$A94D;
	procedure InsertMenuItem (theMenu: MenuHandle;
									itemString: ConstStr255Param;
									afterItem: INTEGER);
	inline
		$A826;
	procedure DeleteMenuItem (theMenu: MenuHandle;
									item: INTEGER);
	inline
		$A952;
	procedure SetMenuItemText (theMenu: MenuHandle;
									item: INTEGER;
									itemString: ConstStr255Param);
	inline
		$A947;
	procedure GetMenuItemText (theMenu: MenuHandle;
									item: INTEGER;
									var itemString: Str255);
	inline
		$A946;
	function GetMenuHandle (menuID: INTEGER): MenuHandle;
	inline
		$A949;
	procedure DeleteMCEntries (menuID: INTEGER;
									menuItem: INTEGER);
	inline
		$AA60;
	procedure DisposeMCInfo (menuCTbl: MCTableHandle);
	inline
		$AA63;
 {OSUtils}
	type
		LongDateRec = DateTimeRec;
		LongDateTime = Longint;
	procedure LongDateToSeconds ({CONST}
									var lDate: LongDateRec;
									var lSecs: LongDateTime);
	inline
		$2F3C, $8008, $FFF2, $A8B5;
	procedure LongSecondsToDate (var lSecs: LongDateTime;
									var lDate: LongDateRec);
	inline
		$2F3C, $8008, $FFF0, $A8B5;
	function IsMetric: BOOLEAN;
	inline
		$3F3C, $0004, $A9ED;
	procedure DateToSeconds ({CONST}
									var d: DateTimeRec;
									var secs: LONGINT);
	procedure SecondsToDate (secs: LONGINT;
									var d: DateTimeRec);
	inline
		$205F, $201F, $A9C6;
{PictUtils}
	function DisposePictInfo (thePictInfoID: Longint): OSErr;
	inline
		$303C, $0206, $A831;
{QuickDraw - unnecessary?}
	procedure DisposePixMap (pm: PixMapHandle);
	inline
		$AA04;
	procedure DisposePixPat (pp: PixPatHandle);
	inline
		$AA08;
	procedure DisposeCTable (cTable: CTabHandle);
	inline
		$AA24;
	procedure DisposeCCursor (cCrsr: CCrsrHandle);
	inline
		$AA26;
	procedure DisposeCIcon (theIcon: CIconHandle);	{Rather Icons.p today?}
	inline
		$AA25;
	procedure DisposeGDevice (gdh: GDHandle);
	inline
		$AA30;
 {QuickdrawText}
{function PixelToChar (textBuf: Ptr; textLength: LONGINT; slop: Fixed; pixelWidth: Fixed; var leadingEdge: BOOLEAN; var widthRemaining: Fixed; styleRunPosition: JustStyleCode; numer: Point; denom: Point): INTEGER;}
{inline}
{$2F3C, $8222, $002E, $A8B5;}
{function CharToPixel (textBuf: Ptr; textLength: LONGINT; slop: Fixed; offset: LONGINT; direction: INTEGER; styleRunPosition: JustStyleCode; numer: Point; denom: Point): INTEGER;}
{inline}
{$2F3C, $821C, $0030, $A8B5;}
{procedure DrawJustified (textPtr: Ptr; textLength: LONGINT; slop: Fixed; styleRunPosition: JustStyleCode; numer: Point; denom: Point);}
{inline}
{$2F3C, $8016, $0032, $A8B5;}
{procedure MeasureJustified (textPtr: Ptr; textLength: LONGINT; slop: Fixed; charLocs: Ptr; styleRunPosition: JustStyleCode; numer: Point; denom: Point);}
{inline}
{$2F3C, $801A, $0034, $A8B5;}
{function PortionLine (textPtr: Ptr; textLen: LONGINT; styleRunPosition: JustStyleCode; numer: Point; denom: Point): Fixed;}
{inline}
{$2F3C, $8412, $0036, $A8B5;}
 {Standard file:}
{Changed call:}
	type
		ConstSFTypeListPtr = ^SFTypeList;
	procedure SFGetFile (where: Point;
									prompt: ConstStr255Param;
									fileFilter: ProcPtr;
									numTypes: INTEGER;
									typeList: ConstSFTypeListPtr;
									dlgHook: ProcPtr;
									var reply: SFReply);
	inline
		$3F3C, $0002, $A9EA;

{Resources}
	function GetResourceSizeOnDisk (theResource: Handle): LONGINT;
	inline
		$A9A5;
	function GetMaxResourceSize (theResource: Handle): LONGINT;
	inline
		$A821;
	procedure RemoveResource (theResource: Handle);
	inline
		$A9AD;

 {TextEdit}
	type
		TEClickLoopUPP = ProcPtr;
		WordBreakUPP = ProcPtr;
		NewTEClickLoopProc = ProcPtr;

	procedure TESetAlignment (just: INTEGER;
									hTE: TEHandle);
	inline
		$A9DF;
	procedure TETextBox (text: univ Ptr;
									length: LONGINT; {CONST}
									var box: Rect;
									just: INTEGER);
	inline
		$A9CE;
	function TEStyleNew ({CONST}
									var destRect: Rect; {CONST}
									var viewRect: Rect): TEHandle;
	inline
		$A83E;
	procedure TESetStyleHandle (theHandle: TEStyleHandle;
									hTE: TEHandle);
	inline
		$3F3C, $0005, $A83D;
	function TEGetStyleHandle (hTE: TEHandle): TEStyleHandle;
	inline
		$3F3C, $0004, $A83D;
	procedure TEStylePaste (hTE: TEHandle);
	inline
		$3F3C, $0000, $A83D;
	function TEGetStyleScrapHandle (hTE: TEHandle): StScrpHandle;
	inline
		$3F3C, $0006, $A83D;
	procedure TEUseStyleScrap (rangeStart: LONGINT;
									rangeEnd: LONGINT;
									newStyles: StScrpHandle;
									fRedraw: BOOLEAN;
									hTE: TEHandle);
	inline
		$3F3C, $000B, $A83D;
	procedure TESetClickLoop (clikProc: TEClickLoopUPP;
									hTE: TEHandle);
	function TEGetScrapLength: LONGINT;
	procedure TESetWordBreak (wBrkProc: WordBreakUPP;
									hTE: TEHandle);
	procedure TEStyleInsert (text: univ Ptr;
									length: LONGINT;
									hST: StScrpHandle;
									hTE: TEHandle);
	inline
		$3F3C, $0007, $A83D;

{ From Dialogs.p }
	function NewModalFilterProc (userRoutine: ModalFilterProcPtr): ModalFilterUPP;
	inline
		$2E9F;
	function GetStdFilterProc (var theProc: ModalFilterUPP): OSErr;
	inline
		$303C, $0203, $AA68;
	function CallModalFilterProc (theDialog: DialogPtr;
									var theEvent: EventRecord;
									var itemHit: Integer;
									userRoutine: ModalFilterUPP): BOOLEAN;
	inline
		$205F, $4E90;
{ From MixedMode.p }
	procedure DisposeRoutineDescriptor (theProcPtr: UniversalProcPtr);
	inline
		$584F;

{$IFC UNDEFINED USE_SANE}
{$SETC USE_SANE = FALSE}
{$ENDC}

{$IFC USE_SANE}
{ Mathematical functions from MathLib that are mapped to SANE }
	function atan (x: Extended): Extended;
{$ENDC}

implementation

{If you need TESetClickLoop, and it isn't implemented by any library you have then you need to include}
{the following code in some unit:}

{procedure TESetClickLoop (clikProc: TEClickLoopUPP; hTE: TEHandle);}
{begin}
{SetClikLoop(clikProc, hTE);}
{end;}

{In SAT, use the file SetClikLoopGlue.p}

{$IFC USE_SANE}
{ Mathematical functions from MathLib that are mapped to SANE }
	function atan (x: Extended): Extended;
	begin
		atan := ArcTan(x);
	end; { atan }
{$ENDC}



{$ifc Use_QDGlobalObjects}
{ Implementation of QuickDraw globals }
	function QDGlobals.randSeed: LONGINT;
		var
			tmp: LongInt;
	begin
		tmp := 1;
		randSeed := tmp;
	end;
	function QDGlobals.screenBits: BitMap;
	begin
		screenBits := screenBits;
	end;
	function QDGlobals.arrow: Cursor;
	begin
		arrow := arrow;
	end;
	function QDGlobals.dkGray: Pattern;
	begin
		dkGray := dkGray;
	end;
	function QDGlobals.ltGray: Pattern;
	begin
		ltGray := ltGray;
	end;
	function QDGlobals.gray: Pattern;
	begin
		gray := gray;
	end;
	function QDGlobals.black: Pattern;
	begin
		black := black;
	end;
	function QDGlobals.white: Pattern;
	begin
		white := white;
	end;
	function QDGlobals.thePort: GrafPtr;
	begin
		thePort := thePort;
	end;
{$endc}

end.