unit Permute;

{ Version 3.4: complete rewrite using TransSkel 2.6 and PowerPC code }
{              Use dynamic memory }

{ Version 3.2 - Multiple Regression tested with permutations, with ter Braak correction }
{ Part of the code is from 'SuperMantel' and 'Procruste'.    }

{ 3.1.1: Program now works correctly with System 6 }
{ 3.2: ter Braak correction (F-tests and t-tests) }

interface

	uses
		MatrixAlgebra, Statistiques;

	const
		PROGRAM_NAME = 'P  E  R  M  U  T  E   !   3 . 4';
		PRINT = True;
		NO_PRINT = False;
		PARTIAL = True;
		NO_PARTIAL = False;

		NO_ALPHA = -1;

		kTestSimple = 1;
		kBackward = 2;
		kForward = 3;
		kStepwise = 4;

	type
		MemPermuteVect = record
				eg: array[1..MAX_DESC] of Integer;
				pg: array[1..MAX_DESC] of Integer;
				egva: array[1..MAX_DESC] of Integer;
				pgva: array[1..MAX_DESC] of Integer;
				prob: array[1..MAX_DESC] of Real;
				probBonf: array[1..MAX_DESC] of Real;
				rSq: Double;
				egr, pgr: Integer;
			end;
		MemPermuteVectPtr = ^MemPermuteVect;
		TPermutation = (vector, matrix, dpt, tpt);
		TTest = (simple, backward, forward, stepwise);

	var
		yFile, xFile: Text;
{$ifc false}
		X: MatrixPtr;
		Y, additiveY, ultrametricY, starY: VectorPtr;
		B: SmallVectorPtr;
		addVect: MemPermuteVect;
		pToEnter, pToRemove: Double;
		xyz: Real;
		varSelectionnees: ICReply; {Sur quelles variables independantes va-t-on faire la regression?}
		gPrintedNotes, gDataStandardized: Boolean;
		gRoot: Byte;
		indVect: VectorIndPtr;
		rootVectY: ObjectVectorPtr;
		rootVectInd: VectorIndPtr;

		varQuelconque: Integer;

	procedure NormalizeVect (var someY: VectorPtr;
									doCenter: Boolean); { Normalize Y }

	procedure NormalizeMat (var someX: MatrixPtr;
									doCenter: Boolean); { Normalize X }

	procedure DoPermTest (Y: VectorPtr;
									X: MatrixPtr;
									var B: SmallVectorPtr;
									var addVect: MemPermuteVect;
									varDepChoisies: ICReply;
									nbPerm, etape: Integer;
									alpha: Double;
									doPrint, isPartial: Boolean;
									theTitle: Str255);

	function SelectSignificanceLevel (theTitle: Str255;
									default: Integer): Double;

	procedure TestQuit (theProc: Str32);
{$endc}

implementation

{$ifc false}
	procedure NormalizeVect (var someY: VectorPtr;
									doCenter: Boolean); { Normalize Y }
		var
			ave, stdev, sum, sumSquares, c: Double;
			i: Integer;
	begin
{ Normalize Y }
		sum := 0;
		sumSquares := 0;
		for i := 1 to nbValeurs do begin
			c := someY^[i];
			sum := sum + c;
			sumSquares := sumSquares + (c * c);
		end;
		stdev := Sqrt((sumSquares - (sum * sum) / nbValeurs) / (nbValeurs - 1));
		if doCenter then begin
			ave := sum / nbValeurs;
			for i := 1 to nbValeurs do
				someY^[i] := (someY^[i] - ave) / stdev;
		end {if doCenter}
		else
			for i := 1 to nbValeurs do
				someY^[i] := someY^[i] / stdev;
	end; { Procedure NormalizeVect }

	procedure NormalizeMat (var someX: MatrixPtr;
									doCenter: Boolean); { Normalize X }
		var
			ave, stdev, sum, sumSquares, c: Double;
			i, j: Integer;
	begin
{ Normalize each column of X }
		for j := 1 to nbDescr do begin
			sum := 0;
			sumSquares := 0;
			for i := 1 to nbValeurs do begin
				c := someX^[i, j];
				sum := sum + c;
				sumSquares := sumSquares + (c * c);
			end;
			stdev := Sqrt((sumSquares - (sum * sum) / nbValeurs) / (nbValeurs - 1));
			if doCenter then begin
				ave := sum / nbValeurs;
				for i := 1 to nbValeurs do
					someX^[i, j] := (someX^[i, j] - ave) / stdev;
			end
			else
				for i := 1 to nbValeurs do
					someX^[i, j] := X^[i, j] / stdev;
		end;
	end; { Procedure NormalizeMat }

	procedure Melange (var theIndVect: VectorIndPtr;
									nombre, noMove: Integer);
		type
			RealArray = array[1..MAX_OBJ] of Real;
		var
			randVect: RealArray;
			s: Integer;

		procedure SHELLSORT (var A_SHELLSORT: RealArray;
										var B_SHELLSORT: VectorIndPtr;
										N: Integer);
			label
				1;
			type
				RealArray = array[1..MAX_OBJ] of Real;
			var
				I, J, K, L, M, O, W, P: INTEGER;
				X: REAL;
		begin
			M := (2 * N - 1) div 2;
			while M <> 0 do begin
				K := N - M;
				for J := 1 to K do begin
					I := J;
					while I >= 1 do begin
						O := I + M;
						P := I;
						if A_SHELLSORT[O] >= A_SHELLSORT[P] then
							goto 1;
						X := A_SHELLSORT[P];
						A_SHELLSORT[P] := A_SHELLSORT[O];
						A_SHELLSORT[O] := X;
						W := B_SHELLSORT^[P];
						B_SHELLSORT^[P] := B_SHELLSORT^[O];
						B_SHELLSORT^[O] := W;
						I := I - M;
					end;
1:
				end;
				M := M div 2;
			end;
		end;        (* FIN DE SHELLSORT *)

	begin { Melange }

		for s := 1 to nombre do
			randVect[s] := Ran3(varQuelconque);
		SHELLSORT(randVect, theIndVect, nombre);
		if noMove > 0 then
			for s := 1 to nombre do
				if theIndVect^[s] = noMove then begin
					theIndVect^[s] := theIndVect^[noMove];
					theIndVect^[noMove] := noMove;
					Leave;
				end;{if theIndVect^[s] = noMove}
	end;{Melange}

	procedure MelangeBig (var theIndVect: BigVectorIndPtr;
									nombre, noMove: Integer);
		type
			RealArray = array[1..MAX_OBJ] of Real;
		var
			randVect: RealArray;
			s: Integer;

		procedure SHELLSORT (var A_SHELLSORT: RealArray;
										var B_SHELLSORT: BigVectorIndPtr;
										N: Integer);
			label
				1;
			type
				RealArray = array[1..MAX_OBJ] of Real;
			var
				I, J, K, L, M, O, W, P: INTEGER;
				XX: REAL;
		begin
			M := (2 * N - 1) div 2;
			while M <> 0 do begin
				K := N - M;
				for J := 1 to K do begin
					I := J;
					while I >= 1 do begin
						O := I + M;
						P := I;
						if A_SHELLSORT[O] >= A_SHELLSORT[P] then
							goto 1;
						XX := A_SHELLSORT[P];
						A_SHELLSORT[P] := A_SHELLSORT[O];
						A_SHELLSORT[O] := XX;
						W := B_SHELLSORT^[P];
						B_SHELLSORT^[P] := B_SHELLSORT^[O];
						B_SHELLSORT^[O] := W;
						I := I - M;
					end;
1:
				end;
				M := M div 2;
			end;
		end;        (* FIN DE SHELLSORT *)

	begin { MelangeBig }

		for s := 1 to nombre do
			randVect[s] := Ran3(varQuelconque);
		SHELLSORT(randVect, theIndVect, nombre);
		if noMove > 0 then
			for s := 1 to nombre do
				if theIndVect^[s] = noMove then begin
					theIndVect^[s] := theIndVect^[noMove];
					theIndVect^[noMove] := noMove;
					Leave;
				end;{if theIndVect^[s] = noMove}
	end;{MelangeBig}

{ procedure DoSimplePerm (var theYVect: VectorPtr);}
{ begin}
{ end;{procedure DoSimplePerm]}
{}
	procedure DoDoublePerm (var theYVect: VectorPtr);
		var
			i, j: Integer;
			offDiag: array[1..MAX_OBJ] of double;
	begin
		Melange(indVect, nbObjets - 1, 0);

{ Cette ligne de code pourrait etre deplace ailleurs pour des raisons d'optimisation }
		for i := 1 to nbObjets - 1 do
			offDiag[i] := ultrametricY^[MtS(i, i + 1)];
{ Il y a peut-etre une erreur a ce niveau: dans le cas de TPT, on passe dans }
{ 'theYVect' le vecteur 'ultrametricY', alors il est modifie a chaque }
{ permutation.  Je me demande si c'est bien raisonnable... }

		for i := 1 to nbObjets - 1 do
			theYVect^[MtS(i, i + 1)] := offDiag[indVect^[i]];
{ J'ai maintenant dans la sur-diagonale de theYVect les valeurs permutees de la }
{ sur-diagonale de la matrice de donnees ultrametriques ultrametricY.  Il faut donc re-remplir }
{ la matrice theYVect en tenant compte de l'inegalite triangulaire. }
{ Ce code provient directement de l'article "The Generation of Random Ultrametric }
{ Matrices Representing Dendrograms", Lapointe & Legendre, J. Class. 8:177-200 (1991). }
{ Il a ete seulement adapte pour tenir compte des structures de donnees en vecteur de }
{ ce programme. }
		for i := 1 to nbObjets - 2 do
			for j := i + 2 to nbObjets do
				if theYVect^[MtS(i, j - 1)] > theYVect^[MtS(j - 1, j)] then
					theYVect^[MtS(i, j)] := theYVect^[MtS(i, j - 1)]
				else
					theYVect^[MtS(i, j)] := theYVect^[MtS(j - 1, j)];
{ Enfin, il faut maintenant normaliser les donnees "brutes" permutees }
	end;{procedure DoDoublePerm}

	procedure DoTriplePerm (var theYVect: VectorPtr);
		var
			i, j, indice: Integer;
	begin
{ Step 1: Permute the root vector }
		Melange(rootVectInd, nbObjets, gRoot);

{ Step 2: Build the 'C' (star tree) vector: }
		indice := 0;
		for i := 1 to nbObjets - 1 do
			for j := i + 1 to nbObjets do begin
				indice := indice + 1;
				starY^[indice] := rootVectY^[rootVectInd^[i]] + rootVectY^[rootVectInd^[j]];
			end;{for j}
{ Step 3: Extract the 'U' (ultrametric tree) vector: }
		for i := 1 to nbValeurs do
			ultrametricY^[i] := additiveY^[i] - starY^[i];

{ Step 4: Permute the 'U' vector: }
		DoDoublePerm(ultrametricY);

{ Step 5: Add 'U' and 'C' to create the permuted 'A' (additive tree) vector: }
		for i := 1 to nbValeurs do
			theYVect^[i] := ultrametricY^[i] - starY^[i];
	end;{procedure DoTriplePerm}

	procedure DoPermTest (Y: VectorPtr;
									X: MatrixPtr;
									var B: SmallVectorPtr;
									var addVect: MemPermuteVect;
									varDepChoisies: ICReply;
									nbPerm, etape: Integer;
									alpha: Double;
									doPrint, isPartial: Boolean;
									theTitle: Str255);
		var
			i, j, petitNbDescr, nbDecimales: Integer;
			zzz, rSqTemp: Double;
			petitX: MatrixPtr;
			tB, BTemp, tBTemp: SmallVectorPtr;
			xPrimeX: SqMatrixPtr;
			correct: Boolean;
			tempStr: Str255;

		procedure CalculerB (var unB: SmallVectorPtr);
			var
				xPrimeY: SmallVectorPtr;
		begin { Procedure CalculerB }
			xPrimeY := SmallVectorPtr(NewPtr(SizeOf(SmallVector)));
			MtVMult(petitX, Y, xPrimeY, nbValeurs, petitNbDescr);{Calculer X'Y}
			MVMult(xPrimeX, xPrimeY, unB, petitNbDescr, petitNbDescr);
			Dispose(xPrimeY);
		end; { Procedure CalculerB }

		function CalculerRCarre (unB: SmallVectorPtr): Double;
			var
				rXY: SmallVectorPtr;
				i, j, tempI, tempJ, tempIndice: Integer;
				sum: Double;
{Note: yBarre = 0, xBarre[i] = 0 car on a centre et reduit}
{  De meme, les ecarts-types sur y et sur x sont egaux a 1}
		begin { Procedure CalculerRCarre }
			rXY := SmallVectorPtr(NewPtr(SizeOf(SmallVector)));
			for i := 1 to petitNbDescr do begin
				sum := 0;
				for j := 1 to nbValeurs do
					if gPermutationalMethod <> kVectorPerm then begin
						tempIndice := j;
						StM(tempIndice, tempI, tempJ);
						tempIndice := MtS(yInd^[tempI], yInd^[tempJ]);
						sum := sum + (Y^[tempIndice] * petitX^[j, i]);
					end { if }
					else { Permutations "normales", i.e. pas Mantel }
						begin
						tempIndice := yIndBig^[j];
						sum := sum + (Y^[tempIndice] * petitX^[j, i]);
					end; { else }
				rXY^[i] := sum / (nbValeurs - 1);
			end;{for i}
			sum := 0;
			for i := 1 to petitNbDescr do begin
				sum := sum + (unB^[i] * rXY^[i]);
			end;{for i}
			Dispose(rXY);
			CalculerRCarre := sum;
		end; { Procedure CalculerRCarre }

	begin
		petitX := MatrixPtr(NewPtr(SizeOf(Matrix)));
		BTemp := SmallVectorPtr(NewPtr(SizeOf(SmallVector)));
		tBTemp := SmallVectorPtr(NewPtr(SizeOf(SmallVector))); { Vecteur de pseudo-t temporaires }
		tB := SmallVectorPtr(NewPtr(SizeOf(SmallVector)));	   { Vecteur de pseudo-t, car on ne calcule que les termes variants }
		xPrimeX := SqMatrixPtr(NewPtr(SizeOf(SqMatrix)));

{ Nous allons utiliser une autre matrice X, soit petitX, pour faire les calculs }
{ en cas de regression partielle }
{ Attention!  Dans ce cas particulier, je lis en colonne (et non pas en ligne) }
{ pour pouvoir compter le nombre de colonnes de X (petitNbDescr) comprises dans la regression!!! }
		petitNbDescr := 0;
		for i := 1 to nbDescr do
			if varDepChoisies[i] then begin
				petitNbDescr := petitNbDescr + 1;
				for j := 1 to nbValeurs do
					petitX^[j, petitNbDescr] := X^[j, i];
			end;

		for i := 1 to petitNbDescr do
			with addVect do begin
				eg[i] := 1;
				egva[i] := 1;
				pg[i] := 0;
				pgva[i] := 0;
				prob[i] := 1;
				probBonf[i] := petitNbDescr;
				pgr := 0;
				egr := 1;
{     pgdr := 0;}
{     egdr := 1;}
			end;

		if gPermutationalMethod <> kVectorPerm then begin
			yInd := VectorIndPtr(NewPtr(SizeOf(VectorInd))); { Nouveau vecteur d'indirections }
			for i := 1 to nbObjets do {Initialiser le vecteur d'indirection}
				yInd^[i] := i;
		end { if }
		else begin
			yIndBig := BigVectorIndPtr(NewPtr(SizeOf(BigVectorInd))); { Nouveau vecteur d'indirections }
			for i := 1 to nbValeurs do {Initialiser le vecteur d'indirection}
				yIndBig^[i] := i;
		end; { else }
{                 Calculer une fois (X'X)^-1                  }
		MtMMult(petitX, xPrimeX, nbValeurs, petitNbDescr);
		correct := MInverse(xPrimeX, petitNbDescr);
		if not correct then begin
			WriteLn(outputFile, 'Matrix X''X is singular (nul determinant).');
			WriteLn(outputFile, 'Program stopped.');
			Quit;
		end;

		if gDebugInfo then begin
{ Sortir la valeur de la diagonale de (X'X)^-1 }
			WriteLn(outputFile, CHR(13), 'Valeur de la diagonale de la matrice (X''X)^-1');
			WriteLn(outputFile, ' i   (X''X)^-1[i,i]   Sqrt((X''X)^-1)');
			for i := 1 to petitNbDescr do
				WriteLn(outputFile, i : 2, '     ', xPrimeX^[i, i] : 12 : 8, '      ', Sqrt(xPrimeX^[i, i]) : 12 : 8);
		end; {Debug Info }

		CalculerB(B); { Calculer le vecteur B original }
		addVect.rSq := CalculerRCarre(B);{ Calculer le R^2 original }
		for j := 1 to nbDescr do
{ Calculer le vecteur de pseudo-t s a partir du vecteur des b s }
{ car pseudo-t = b / SQRT(1-R^2), cf S&R page 636 point 4 }
			begin
			tB^[j] := B^[j] / Sqrt(1 - addVect.rSq);
		end; { for j }

		if gDebugInfo then begin
{ Sortir le vecteur des pseudo-t s pour des fins de verification }
			WriteLn(outputFile, CHR(13), 'Valeur du vecteur des pseudo-t s [b/Sqrt(1-R^2)]');
			WriteLn(outputFile, ' i      pseudo-t');
			for j := 1 to nbDescr do
				WriteLn(outputFile, j : 2, tB^[j] : 12 : 8);
		end; {Debug Info }

		if nbPerm > 0 then begin
			varQuelconque := -1;
			for i := 1 to 100 do (*Initialiser Ran3*)
				zzz := Ran3(varQuelconque);

			if theTitle = '' then
				if etape > 0 then begin
					NumToString(etape, tempStr);
					tempStr := Concat('Permutations: step ', tempStr);
				end
				else
					tempStr := 'Permutations progress'
			else
				tempStr := theTitle;
			SetupProgress(tempStr, 0, nbPerm);

			if gPermutationalMethod in [kDoublePerm, kTriplePerm] then
				for i := 2 to nbObjets do {'2' car c'est pour la sur-diagonale}
					indVect^[i - 1] := i;

			i := 1;
			while (i <= nbPerm) and (not gQuitNow) do
          {Calcul de la distribution de la statistique permute}
				begin
				case gPermutationalMethod of
					kVectorPerm, kMatrixPerm: 
						;{default mode}
					kDoublePerm:  begin
						DoDoublePerm(Y);
						NormalizeVect(Y, True);
					end;{kDoublePerm}
					kTriplePerm:  begin
						DoTriplePerm(Y);
						NormalizeVect(Y, True);
					end;{kTriplePerm}
				end;{case}

				if gPermutationalMethod <> kVectorPerm then
{Melanger le vecteur d'indirections}
					Melange(yInd, nbObjets, 0)
				else
					MelangeBig(yIndBig, nbObjets, 0);

				CalculerB(BTemp);
				rSqTemp := CalculerRCarre(BTemp);
				for j := 1 to nbDescr do
{ Calculer le vecteur de pseudo-t s a partir du vecteur des b s }
{ car pseudo-t = b / SQRT(1-R^2), cf S&R page 636 point 4 }
					begin
					tBTemp^[j] := BTemp^[j] / Sqrt(1 - rSqTemp);
				end; { for j }

				for j := 1 to petitNbDescr do begin
{                               Test en tenant compte du signe               }
					if tBTemp^[j] = tB^[j] then
						addVect.eg[j] := addVect.eg[j] + 1
					else if tBTemp^[j] > tB^[j] then
						addVect.pg[j] := addVect.pg[j] + 1;
{                               Test en valeur absolue                       }
					if abs(tBTemp^[j]) = abs(tB^[j]) then
						addVect.egva[j] := addVect.egva[j] + 1
					else if abs(tBTemp^[j]) > abs(tB^[j]) then
						addVect.pgva[j] := addVect.pgva[j] + 1;
				end;{ for j }
{  Meme chose, mais pour le r^2 (sauf qu'il est toujours positif!!           }
				if rSqTemp = addVect.rSq then
					addVect.egr := addVect.egr + 1
				else if rSqTemp > addVect.rSq then
					addVect.pgr := addVect.pgr + 1;

				ProgressInfo(i);
				i := i + 1;
			end; {while (i <= nbPerm) and (not gQuitNow)}
			DisposeProgress(True);

{ Calculer les probabilites unilaterales associees a chaque tB: }
{ (on laisse les Bonferonni pour plus tard...)                 }
			for j := 1 to petitNbDescr do
				if tB^[j] > 0 then { Calcul a droite }
					addVect.prob[j] := (addVect.eg[j] + addVect.pg[j]) / (nbPerm + 1)
				else { Calcul a gauche: Modification PhC 25/04/95 }
					addVect.prob[j] := (nbPerm - addVect.pg[j] + addVect.eg[j]) / (nbPerm + 1);

		end; { if nbPerm > 0 }
		if gQuitNow then begin
			WriteLn(outputFile);
			Write(outputFile, '### COMPUTATIONS HALTED BY USER AT ');
			Write(outputFile, i : 1, ' PERMUTATIONS ###');
			nbPerm := i - 1;
		end;{if gQuitNow}
		if doPrint then begin
			if nbPerm > 0 then
				nbDecimales := Trunc(Ln(nbPerm) / Ln(10)) + 1
			else
				nbDecimales := 3; { valeur par defaut }
			if not gPrintedNotes then begin
				WriteLn(outputFile);
				WriteLn(outputFile, 'NOTES: - EQual values include the initial statistic.');
				WriteLn(outputFile, '       - B coefficients are standardized.');
				WriteLn(outputFile, '       - Probabilities of the B coefficients are actually');
				WriteLn(outputFile, '       -     those of the associated t-statistics.');
				WriteLn(outputFile);
				gPrintedNotes := True;
			end;{if not gPrintedNotes}

{ Imprimer la probabilite de Bonferroni }
			if alpha <> NO_ALPHA then begin
				Write(outputFile, '     (*) Reference: Bonferonni-corrected alpha-level = ');
				WriteLn(outputFile, (alpha / petitNbDescr) : (nbDecimales + 2) : nbDecimales);
				WriteLn(outputFile);
			end; { if alpha <> NO_ALPHA }

			if (petitNbDescr < nbDescr) and (etape = 0) then begin
				Write(outputFile, 'CAUTION! ---> Multiple regression with only ');
				WriteLn(outputFile, petitNbDescr : 1, ' variable(s) out of ', nbDescr : 1);
				WriteLn(outputFile);
			end;
			if etape > 0 then begin
				WriteLn(outputFile, 'Step ', etape : 1, ':');
			end;
			if isPartial then
				WriteLn(outputFile, 'Variable  Partial B. Coeff.  EQ    GT    |EQ|    |GT|  Unil. prob. ')
			else
				WriteLn(outputFile, 'Variable  B Coefficient      EQ    GT    |EQ|    |GT|  Unil. prob.');
			WriteLn(outputFile, '-----------------------------------------------------  -----------');
			j := 0; { Necessaire pour n'afficher que les variables sur lesquelles on a fait l'analyse }
			for i := 1 to nbDescr do begin
				if varDepChoisies[i] then begin
					j := j + 1;
					Write(outputFile, ' X', i : 1, '         ');
					if B^[j] > 0 then
						Write(outputFile, ' ');
					Write(outputFile, B^[j] : 10 : 8, '    ');
					Write(outputFile, addVect.eg[j] : 4, '  ', addVect.pg[j] : 4, '   ');
					Write(outputFile, addVect.egva[j] : 4, '    ', addVect.pgva[j] : 4);
{ Calcul des probabilites de Bonferronni }
					if isPartial then
						addVect.probBonf[j] := addVect.prob[j] * etape
					else
						addVect.probBonf[j] := addVect.prob[j] * petitNbDescr;
					if nbPerm > 0 then
						Write(outputFile, '      ', addVect.prob[j] : (nbDecimales + 2) : nbDecimales);
					WriteLn(outputFile);
				end;{if varDepChoisies[i]}
			end;{for i}
			WriteLn(outputFile);
			if isPartial then
				WriteLn(outputFile, ' Partial R^2  EQ    GT')
			else
				WriteLn(outputFile, '   R^2        EQ    GT   Prob ');
			WriteLn(outputFile, '----------------------  ------');
			Write(outputFile, addVect.rSq : 10 : 8, '  ');
			if nbPerm > 0 then begin
				Write(outputFile, addVect.egr : 4, '  ', addVect.pgr : 4, '  ');
				WriteLn(outputFile, ((addVect.egr + addVect.pgr) / (nbPerm + 1)) : 6 : 4);
			end
			else
				WriteLn(outputFile);
			WriteLn(outputFile);
		end;{if doPrint}

		Dispose(petitX);
		Dispose(BTemp);
		Dispose(tBTemp);
		Dispose(tB);
		Dispose(xPrimeX);
		if gPermutationalMethod <> kVectorPerm then
			Dispose(yInd)
		else
			Dispose(yIndBig);
	end; { Procedure DoPerm }

	procedure TestQuit (theProc: Str32);
	begin
		if gQuitNow then begin
			WriteLn(outputFile);
			WriteLn(outputFile, '### Procedure ', theProc, ' interrupted by user ###');
			WriteLn(outputFile);
			Quit;
		end;{if}
	end;{procedure TestQuit}

	function SelectSignificanceLevel (theTitle: Str255;
									default: Integer): Double;
		var
			i: Integer;
			lesDescr: ChoiceArray;
	begin
		lesDescr[1] := '0.1';
		lesDescr[2] := '0.05';
		lesDescr[3] := '0.01';
		lesDescr[4] := '0.001';
		i := default;

		if theTitle = '' then
			theTitle := 'Select the alpha significance level';

		GetExclusiveChoice(theTitle, 4, lesDescr, i);
		case i of
			1: 
				SelectSignificanceLevel := 0.1;
			2: 
				SelectSignificanceLevel := 0.05;
			3: 
				SelectSignificanceLevel := 0.01;
			4: 
				SelectSignificanceLevel := 0.001;
			otherwise
				Quit;
		end;{case}
	end;{function SelectSignificanceLevel}
{$endc}
end. { unit Permute }